#  *******************************************************************************
#  Copyright (c) 2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import pytest

from otterdog.config import OtterdogConfig
from otterdog.webapp import mongo
from otterdog.webapp.db.models import InstallationModel


@pytest.mark.asyncio
@pytest.mark.skip(reason="integration test")
async def test_update_installations(app):
    from otterdog.webapp.db import init_mongo_database
    from otterdog.webapp.db.service import (
        get_all_installations_count,
        get_installations,
        update_installations_from_config,
    )

    async with app.app_context():
        await mongo.odm.get_collection(InstallationModel).drop()
        await init_mongo_database(mongo)

        policies = []

        config = OtterdogConfig.from_file(_get_config_file("config/otterdog.json"), False)
        await update_installations_from_config(config, policies, False)

        assert await get_all_installations_count() == 1
        assert (await get_installations())[0].github_id == "OtterdogTest"

        config2 = OtterdogConfig.from_file(_get_config_file("config/otterdog_modified.json"), False)
        await update_installations_from_config(config2, policies, False)

        assert await get_all_installations_count() == 1
        assert (await get_installations())[0].github_id == "OtterdogTest2"


def _get_config_file(filename: str) -> str:
    import os

    dirname = os.path.dirname(__file__)
    return os.path.join(dirname, filename)



#  *******************************************************************************
#  Copyright (c) 2023-2025 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING, Protocol

import aiofiles.ospath

from otterdog.models import LivePatch, LivePatchContext, LivePatchType
from otterdog.models.github_organization import GitHubOrganization
from otterdog.providers.github import GitHubProvider
from otterdog.utils import Change, IndentingPrinter, unwrap

from . import Operation
from .validate import ValidateOperation

if TYPE_CHECKING:
    from typing import Any

    from otterdog.config import OrganizationConfig, OtterdogConfig
    from otterdog.jsonnet import JsonnetConfig
    from otterdog.models import ModelObject


class DiffStatus:
    def __init__(self):
        self.additions = 0
        self.differences = 0
        self.deletions = 0

    def total_changes(self, include_deletions: bool) -> int:
        if include_deletions:
            return self.additions + self.differences + self.deletions
        else:
            return self.additions + self.differences


class CallbackFn(Protocol):
    def __call__(self, org_id: str, diff_status: DiffStatus, patches: list[LivePatch]) -> None: ...


class DiffOperation(Operation):
    def __init__(
        self,
        no_web_ui: bool,
        repo_filter: str,
        update_webhooks: bool,
        update_secrets: bool,
        update_filter: str,
    ):
        super().__init__()

        self.no_web_ui = no_web_ui
        self.repo_filter = repo_filter
        self.update_webhooks = update_webhooks
        self.update_secrets = update_secrets
        self.update_filter = update_filter
        self._gh_client: GitHubProvider | None = None
        self._validator = ValidateOperation()
        self._template_dir: str | None = None
        self._org_config: OrganizationConfig | None = None
        self._callback: CallbackFn | None = None
        self._concurrency: int | None = None

    @property
    def template_dir(self) -> str:
        return unwrap(self._template_dir)

    @property
    def org_config(self) -> OrganizationConfig:
        return unwrap(self._org_config)

    @property
    def concurrency(self) -> int | None:
        return self._concurrency

    @concurrency.setter
    def concurrency(self, value: int) -> None:
        self._concurrency = value

    def set_callback(self, fn: CallbackFn) -> None:
        self._callback = fn

    def init(self, config: OtterdogConfig, printer: IndentingPrinter) -> None:
        super().init(config, printer)
        self._validator.init(config, printer)

    async def execute(
        self,
        org_config: OrganizationConfig,
        org_index: int | None = None,
        org_count: int | None = None,
    ) -> int:
        self._org_config = org_config

        self._print_project_header(org_config, org_index, org_count)
        self.printer.level_up()

        try:
            return await self.generate_diff(org_config)
        finally:
            self.printer.level_down()

    async def generate_diff(self, org_config: OrganizationConfig) -> int:
        try:
            self._gh_client = self.setup_github_client(org_config)
        except RuntimeError as e:
            self.printer.print_error(f"invalid credentials\n{e!s}")
            return 1

        try:
            return await self._generate_diff_internal(org_config)
        except RuntimeError as e:
            self.printer.print_error(f"planning aborted: {e!s}")
            return 1
        finally:
            if self._gh_client is not None:
                await self._gh_client.close()

    def setup_github_client(self, org_config: OrganizationConfig) -> GitHubProvider:
        return GitHubProvider(self.get_credentials(org_config, only_token=self.no_web_ui))

    @property
    def gh_client(self) -> GitHubProvider:
        return unwrap(self._gh_client)

    def verbose_output(self):
        return True

    def include_resources_with_secrets(self) -> bool:
        return True

    def resolve_secrets(self) -> bool:
        return True

    async def _generate_diff_internal(self, org_config: OrganizationConfig) -> int:
        github_id = org_config.github_id
        jsonnet_config = org_config.jsonnet_config
        await jsonnet_config.init_template()
        self._template_dir = jsonnet_config.template_dir

        org_file_name = jsonnet_config.org_config_file

        if not await aiofiles.ospath.exists(org_file_name):
            self.printer.print_error(
                f"configuration file '{org_file_name}' does not yet exist, run fetch-config or import first."
            )
            return 1

        try:
            expected_org = self.load_expected_org(github_id, org_file_name)
        except RuntimeError as e:
            self.printer.print_error(f"failed to load configuration\n{e!s}")
            return 1

        # We validate the configuration first and only calculate a plan if
        # there are no validation errors.
        (
            validation_infos,
            validation_warnings,
            validation_errors,
        ) = await self._validator.validate(expected_org, jsonnet_config, self.gh_client)
        if validation_errors > 0:
            self.printer.println("Planning aborted due to validation errors.")
            return validation_errors

        if validation_infos > 0 and not self.printer.is_info_enabled():
            self.printer.println(
                f"there have been {validation_infos} validation infos, enable verbose output to display them."
            )

        try:
            current_org = await self.load_current_org(org_config.name, github_id, jsonnet_config)
        except RuntimeError as e:
            self.printer.print_error(f"failed to load current configuration\n{e!s}")
            return 1

        expected_org, current_org = self.preprocess_orgs(expected_org, current_org)

        diff_status = DiffStatus()
        live_patches = []

        def handle(patch: LivePatch) -> None:
            if not self.include_resources_with_secrets() and patch.requires_secrets():
                return

            live_patches.append(patch)

            match patch.patch_type:
                case LivePatchType.ADD:
                    self.handle_add_object(github_id, unwrap(patch.expected_object), patch.parent_object)
                    diff_status.additions += 1

                case LivePatchType.REMOVE:
                    self.handle_delete_object(github_id, unwrap(patch.current_object), patch.parent_object)
                    diff_status.deletions += 1

                case LivePatchType.CHANGE:
                    diff_status.differences += self.handle_modified_object(
                        github_id,
                        unwrap(patch.changes),
                        patch.forced_update,
                        unwrap(patch.current_object),
                        unwrap(patch.expected_object),
                        patch.parent_object,
                    )

        context = LivePatchContext(
            github_id,
            self.repo_filter,
            self.update_webhooks,
            self.update_secrets,
            self.update_filter,
            current_org.settings if self.coerce_current_org() else None,
            expected_org.settings,
        )
        expected_org.generate_live_patch(current_org, context, handle)

        # resolve secrets for collected patches
        if self.resolve_secrets():
            for live_patch in live_patches:
                if live_patch.expected_object is not None:
                    live_patch.expected_object.resolve_secrets(self.credential_resolver.get_secret)

        status = await self.handle_finish(github_id, diff_status, live_patches)

        if self._callback is not None:
            self._callback(github_id, diff_status, live_patches)

        return status

    def load_expected_org(self, github_id: str, org_file_name: str) -> GitHubOrganization:
        return GitHubOrganization.load_from_file(github_id, org_file_name)

    def coerce_current_org(self) -> bool:
        return False

    async def load_current_org(
        self, project_name: str, github_id: str, jsonnet_config: JsonnetConfig
    ) -> GitHubOrganization:
        return await GitHubOrganization.load_from_provider(
            project_name,
            github_id,
            jsonnet_config,
            self.gh_client,
            self.no_web_ui,
            self.concurrency,
            self.repo_filter,
            exclude_teams=self.config.exclude_teams_pattern,
        )

    def preprocess_orgs(
        self, expected_org: GitHubOrganization, current_org: GitHubOrganization
    ) -> tuple[GitHubOrganization, GitHubOrganization]:
        return expected_org, current_org

    @abstractmethod
    def handle_add_object(
        self,
        org_id: str,
        model_object: ModelObject,
        parent_object: ModelObject | None = None,
    ) -> None: ...

    @abstractmethod
    def handle_delete_object(
        self,
        org_id: str,
        model_object: ModelObject,
        parent_object: ModelObject | None = None,
    ) -> None: ...

    @abstractmethod
    def handle_modified_object(
        self,
        org_id: str,
        modified_object: dict[str, Change[Any]],
        forced_update: bool,
        current_object: ModelObject,
        expected_object: ModelObject,
        parent_object: ModelObject | None = None,
    ) -> int: ...

    @abstractmethod
    async def handle_finish(self, org_id: str, diff_status: DiffStatus, patches: list[LivePatch]) -> int: ...



#  *******************************************************************************
#  Copyright (c) 2023-2025 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from collections.abc import Mapping
from typing import Any

from otterdog.jsonnet import JsonnetConfig
from otterdog.models import ModelObject
from otterdog.models.organization_settings import OrganizationSettings
from otterdog.utils import UNSET, Change

from . import ModelTest


class OrganizationSettingsTest(ModelTest):
    def create_model(self, data: Mapping[str, Any]) -> ModelObject:
        return OrganizationSettings.from_model_data(data["settings"])

    @property
    def template_function(self) -> str:
        return JsonnetConfig.create_org

    @property
    def model_data(self):
        return self.load_json_resource("otterdog-org-settings.json")

    @property
    def provider_data(self):
        return self.load_json_resource("github-org-settings.json")

    def test_load_from_model(self):
        settings = OrganizationSettings.from_model_data(self.model_data)

        assert settings.name is None
        assert settings.plan == "free"
        assert settings.description is None
        assert settings.email is None
        assert settings.location is None
        assert settings.company is None
        assert settings.billing_email == "thomas.neidhart@eclipse-foundation.org"
        assert settings.twitter_username is None
        assert settings.blog is None
        assert settings.has_organization_projects is True
        assert settings.default_branch_name == "main"
        assert settings.default_repository_permission == "read"
        assert settings.two_factor_requirement is False
        assert settings.web_commit_signoff_required is False
        assert settings.members_can_create_private_repositories is False
        assert settings.members_can_create_public_repositories is True
        assert settings.members_can_fork_private_repositories is True
        assert settings.members_can_create_public_pages is True
        assert settings.members_can_change_repo_visibility is False
        assert settings.members_can_delete_repositories is True
        assert settings.members_can_delete_issues is False
        assert settings.members_can_create_teams is True
        assert settings.readers_can_create_discussions is False
        assert settings.packages_containers_public is False
        assert settings.packages_containers_internal is False
        assert settings.members_can_change_project_visibility is False

    def test_load_from_provider(self):
        settings = OrganizationSettings.from_provider_data(self.org_id, self.provider_data)

        assert settings.name is None
        assert settings.plan == "free"
        assert settings.description is None
        assert settings.email is None
        assert settings.location is None
        assert settings.company is None
        assert settings.billing_email == "thomas.neidhart@eclipse-foundation.org"
        assert settings.twitter_username is None
        assert settings.blog is None
        assert settings.has_organization_projects is True
        assert settings.default_branch_name == "main"
        assert settings.two_factor_requirement is False
        assert settings.web_commit_signoff_required is False
        assert settings.members_can_create_private_repositories is False
        assert settings.members_can_create_public_repositories is True
        assert settings.members_can_fork_private_repositories is True
        assert settings.members_can_create_public_pages is True
        assert settings.members_can_change_repo_visibility is False
        assert settings.members_can_delete_repositories is True
        assert settings.members_can_delete_issues is False
        assert settings.members_can_create_teams is True
        assert settings.readers_can_create_discussions is False
        assert settings.packages_containers_public is False
        assert settings.packages_containers_internal is False
        assert settings.members_can_change_project_visibility is False

    async def test_to_provider(self):
        settings = OrganizationSettings.from_model_data(self.model_data)

        settings.description = UNSET

        provider_data = await settings.to_provider_data(self.org_id, self.provider)

        assert len(provider_data) == 23
        assert provider_data["billing_email"] == settings.billing_email

    async def test_changes_to_provider(self):
        current = OrganizationSettings.from_model_data(self.model_data)
        other = OrganizationSettings.from_model_data(self.model_data)

        other.billing_email = "mikael_barbero@eclipse-foundation.org"
        other.default_repository_permission = "none"

        changes = current.get_difference_from(other)
        provider_data = await OrganizationSettings.changes_to_provider(self.org_id, changes, self.provider)

        assert len(provider_data) == 2
        assert provider_data["billing_email"] == current.billing_email
        assert provider_data["default_repository_permission"] == current.default_repository_permission

    def test_patch(self):
        current = OrganizationSettings.from_model_data(self.model_data)

        default = OrganizationSettings.from_model_data(self.model_data)

        default.billing_email = None
        default.web_commit_signoff_required = True

        patch = current.get_patch_to(default)

        assert len(patch) == 2
        assert patch["billing_email"] == current.billing_email
        assert patch["web_commit_signoff_required"] is current.web_commit_signoff_required

    def test_difference(self):
        current = OrganizationSettings.from_model_data(self.model_data)
        other = OrganizationSettings.from_model_data(self.model_data)

        other.billing_email = "mikael_barbero@eclipse-foundation.org"
        other.default_repository_permission = "none"

        diff = current.get_difference_from(other)

        assert len(diff) == 2
        assert diff["billing_email"] == Change(other.billing_email, current.billing_email)
        assert diff["default_repository_permission"] == Change(
            other.default_repository_permission, current.default_repository_permission
        )



#  *******************************************************************************
#  Copyright (c) 2023-2025 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from collections.abc import Mapping
from typing import Any

from otterdog.jsonnet import JsonnetConfig
from otterdog.models import ModelObject
from otterdog.models.organization_ruleset import OrganizationRuleset

from . import ModelTest


class OrganizationRulesetTest(ModelTest):
    def create_model(self, data: Mapping[str, Any]) -> ModelObject:
        return OrganizationRuleset.from_model_data(data)

    @property
    def template_function(self) -> str:
        return JsonnetConfig.create_org_ruleset

    @property
    def model_data(self):
        raise NotImplementedError

    @property
    def provider_data(self):
        raise NotImplementedError



#  *******************************************************************************
#  Copyright (c) 2023-2025 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import asyncio
import json
import os
import pathlib
import re
import tempfile
import zipfile
from typing import Any

import aiofiles
import chevron

from otterdog.logging import is_trace_enabled
from otterdog.providers.github.exception import GitHubException
from otterdog.providers.github.rest import RestApi, RestClient, encrypt_value
from otterdog.utils import (
    associate_by_key,
    get_logger,
    is_set_and_present,
    query_json,
)

_logger = get_logger(__name__)


class RepoClient(RestClient):
    def __init__(self, rest_api: RestApi):
        super().__init__(rest_api)

    async def get_simple_repo_data(self, org_id: str, repo_name: str) -> dict[str, Any]:
        _logger.debug("retrieving simple repo data for '%s/%s'", org_id, repo_name)

        try:
            return await self.requester.request_json("GET", f"/repos/{org_id}/{repo_name}")
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving simple repo data for repo '{org_id}/{repo_name}':\n{ex}") from ex

    async def get_default_branch(self, org_id: str, repo_name: str) -> str:
        _logger.debug("retrieving default branch for repo '%s/%s'", org_id, repo_name)

        repo_data = await self.get_simple_repo_data(org_id, repo_name)
        return repo_data["default_branch"]

    async def get_branch(self, org_id: str, repo_name: str, branch_name: str) -> dict[str, Any]:
        _logger.debug("retrieving data for branch '%s' in repo '%s/%s'", branch_name, org_id, repo_name)

        try:
            return await self.requester.request_json("GET", f"/repos/{org_id}/{repo_name}/branches/{branch_name}")
        except GitHubException as ex:
            raise RuntimeError(
                f"failed retrieving data for branch '{branch_name}' in repo '{org_id}/{repo_name}':\n{ex}"
            ) from ex

    async def rename_branch(self, org_id: str, repo_name: str, branch: str, new_name: str) -> dict[str, Any]:
        _logger.debug("renaming branch '%s' to '%s' in repo '%s/%s'", branch, new_name, org_id, repo_name)

        try:
            data = {"new_name": new_name}
            return await self.requester.request_json(
                "POST", f"/repos/{org_id}/{repo_name}/branches/{branch}/rename", data
            )
        except GitHubException as ex:
            raise RuntimeError(f"failed renaming branch '{branch}' in repo '{org_id}/{repo_name}':\n{ex}") from ex

    async def get_repo_data(self, org_id: str, repo_name: str) -> dict[str, Any]:
        _logger.debug("retrieving repo data for '%s/%s'", org_id, repo_name)

        try:
            repo_data = await self.get_simple_repo_data(org_id, repo_name)

            archived = repo_data.get("archived", False)
            if not archived:
                await self._fill_vulnerability_alerts(org_id, repo_name, repo_data)

                private = repo_data.get("private", False)
                if not private:
                    await self._fill_private_vulnerability_reporting(org_id, repo_name, repo_data)

            await self._fill_github_pages_config(org_id, repo_name, repo_data)
            await self._fill_topics(org_id, repo_name, repo_data)
            await self._fill_code_scanning_config(org_id, repo_name, repo_data)
            await self._fill_custom_properties(org_id, repo_name, repo_data)

            return repo_data
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving data for repo '{repo_name}':\n{ex}") from ex

    async def get_repo_by_id(self, repo_id: int) -> dict[str, Any]:
        _logger.debug("retrieving repo by id for '%d'", repo_id)

        try:
            return await self.requester.request_json("GET", f"/repositories/{repo_id}")
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving data for repo '{repo_id}':\n{ex}") from ex

    async def update_repo(self, org_id: str, repo_name: str, data: dict[str, Any]) -> None:
        _logger.debug("updating repo settings for repo '%s/%s'", org_id, repo_name)

        changes = len(data)

        if "dependabot_alerts_enabled" in data:
            vulnerability_alerts = bool(data.pop("dependabot_alerts_enabled"))
        else:
            vulnerability_alerts = None

        if "private_vulnerability_reporting_enabled" in data:
            private_vulnerability_reporting = bool(data.pop("private_vulnerability_reporting_enabled"))
        else:
            private_vulnerability_reporting = None

        topics = list(data.pop("topics")) if "topics" in data else None
        custom_properties = list(data.pop("custom_properties")) if "custom_properties" in data else None
        gh_pages = data.pop("gh_pages") if "gh_pages" in data else None
        code_scanning = data.pop("code_scanning_default_config") if "code_scanning_default_config" in data else None
        default_branch = data.pop("default_branch") if "default_branch" in data else None

        if changes > 0:
            try:
                if len(data) > 0:
                    await self.requester.request_json("PATCH", f"/repos/{org_id}/{repo_name}", data)

                if vulnerability_alerts is not None:
                    await self._update_vulnerability_alerts(org_id, repo_name, vulnerability_alerts)
                if private_vulnerability_reporting is not None:
                    await self._update_private_vulnerability_reporting(
                        org_id, repo_name, private_vulnerability_reporting
                    )
                if topics is not None:
                    await self._update_topics(org_id, repo_name, topics)
                if custom_properties is not None and len(custom_properties) > 0:
                    await self._update_custom_properties(org_id, repo_name, custom_properties)
                if gh_pages is not None:
                    await self._update_github_pages_config(org_id, repo_name, gh_pages)
                if code_scanning is not None:
                    await self._update_code_scanning_config(org_id, repo_name, code_scanning)
                if default_branch is not None:
                    await self._update_default_branch(org_id, repo_name, default_branch)

                _logger.debug("updated %d repo setting(s) for repo '%s/%s'", changes, org_id, repo_name)
            except GitHubException as ex:
                raise RuntimeError(f"failed to update settings for repo '{repo_name}':\n{ex}") from ex

    async def add_repo(
        self,
        org_id: str,
        data: dict[str, Any],
        template_repository: str | None,
        post_process_template_content: list[str],
        forked_repository: str | None,
        fork_default_branch_only: bool,
        auto_init_repo: bool,
    ) -> None:
        repo_name = data["name"]

        if is_set_and_present(forked_repository):
            _logger.debug("forking repo '%s' to '%s/%s'", forked_repository, org_id, repo_name)
            upstream_owner, upstream_repo = re.split("/", forked_repository, maxsplit=1)

            try:
                fork_data = {
                    "organization": org_id,
                    "name": repo_name,
                    "default_branch_only": fork_default_branch_only,
                }

                await self.requester.request_json(
                    "POST",
                    f"/repos/{upstream_owner}/{upstream_repo}/forks",
                    fork_data,
                )

                # get all the data for the created repo to avoid setting values that can not be changed due
                # to defaults from the organization (like web_commit_signoff_required)
                current_data = await self.get_repo_data(org_id, repo_name)
                self._remove_already_active_settings(data, current_data)
                await self.update_repo(org_id, repo_name, data)

                _logger.debug("forked repo with name '%s' from repo '%s'", repo_name, forked_repository)
                return
            except GitHubException as ex:
                raise RuntimeError(f"failed to fork repo '{repo_name}' from repo '{forked_repository}':\n{ex}") from ex

        if is_set_and_present(template_repository):
            _logger.debug("creating repo '%s/%s' from template '%s'", org_id, repo_name, template_repository)
            template_owner, template_repo = re.split("/", template_repository, maxsplit=1)

            try:
                template_data = {
                    "owner": org_id,
                    "name": repo_name,
                    "include_all_branches": False,
                    "private": data.get("private", False),
                }

                await self.requester.request_json(
                    "POST",
                    f"/repos/{template_owner}/{template_repo}/generate",
                    template_data,
                )

                _logger.debug("created repo with name '%s' from template '%s'", repo_name, template_repository)

                # get all the data for the created repo to avoid setting values that can not be changed due
                # to defaults from the organization (like web_commit_signoff_required)
                current_data = await self.get_repo_data(org_id, repo_name)
                self._remove_already_active_settings(data, current_data)
                await self.update_repo(org_id, repo_name, data)

                # wait till the repo is initialized, this might take a while.
                if len(post_process_template_content) > 0:
                    initialized = False
                    for i in range(1, 11):
                        try:
                            await self.get_readme(org_id, repo_name)
                            initialized = True
                            break
                        except RuntimeError:
                            _logger.trace(f"waiting for repo '{org_id}/{repo_name}' to be initialized, try {i} of 10")
                            import time

                            time.sleep(1)

                    if initialized is False:
                        raise RuntimeError(
                            f"failed to create repo from template '{template_repository}': "
                            f"repo not initialized after 5s"
                        )

                # if there is template content which shall be post-processed,
                # use chevron to expand some variables that might be used there.
                for content_path in post_process_template_content:
                    content = await self.rest_api.content.get_content(org_id, repo_name, content_path, None)
                    updated_content = self._render_template_content(org_id, repo_name, content)
                    if content != updated_content:
                        await self.rest_api.content.update_content(org_id, repo_name, content_path, updated_content)

                return
            except GitHubException as ex:
                raise RuntimeError(f"failed to create repo from template '{template_repository}':\n{ex}") from ex

        _logger.debug("creating repo '%s/%s'", org_id, repo_name)

        # some settings do not seem to be set correctly during creation
        # collect them and update the repo after creation.
        update_keys = [
            "dependabot_alerts_enabled",
            "web_commit_signoff_required",
            "security_and_analysis",
            "topics",
            "gh_pages",
            "code_scanning_default_config",
            "custom_properties",
            "private_vulnerability_reporting_enabled",
        ]

        if auto_init_repo is True:
            update_keys.append("default_branch")

        update_data = {}

        for update_key in update_keys:
            if update_key in data:
                update_data[update_key] = data.pop(update_key)

        # whether the repo should be initialized with an empty README
        data["auto_init"] = auto_init_repo

        # if gh pages are disabled, do not try to update the config as it will fail
        if "gh_pages" in update_data:
            gh_pages = update_data.get("gh_pages", {})
            build_type = gh_pages.get("build_type")
            if build_type == "disabled":
                update_data.pop("gh_pages")

        try:
            result = await self.requester.request_json("POST", f"/orgs/{org_id}/repos", data)
            _logger.debug("created repo with name '%s'", repo_name)
            self._remove_already_active_settings(update_data, result)
            # let's wait a bit for the repo to be created.
            # in some cases, the repo can not be access yet via the REST API, resulting in 404 errors
            # we prevent that by adding an arbitrary sleep
            await asyncio.sleep(2)
            await self.update_repo(org_id, repo_name, update_data)
        except GitHubException as ex:
            raise RuntimeError(f"failed to add repo with name '{org_id}/{repo_name}':\n{ex}") from ex

    async def get_webhook_id(self, org_id: str, repo_name: str, url: str) -> str:
        _logger.debug("retrieving id for repo webhook with url '%s' for repo '%s/%s'", url, org_id, repo_name)

        webhooks = await self.get_webhooks(org_id, repo_name)

        has_wildcard_url = url.endswith("*")
        stripped_url = url.rstrip("*")

        for webhook in webhooks:
            webhook_url = webhook["config"]["url"]
            if (has_wildcard_url is True and webhook_url.startswith(stripped_url)) or webhook_url == url:
                return webhook["id"]

        raise RuntimeError(f"failed to find repo webhook with url '{url}'")

    async def get_webhooks(self, org_id: str, repo_name: str) -> list[dict[str, Any]]:
        _logger.debug("retrieving webhooks for repo '%s/%s'", org_id, repo_name)

        try:
            return await self.requester.request_json("GET", f"/repos/{org_id}/{repo_name}/hooks")
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving webhooks for repo '{org_id}/{repo_name}':\n{ex}") from ex

    async def update_webhook(self, org_id: str, repo_name: str, webhook_id: int, webhook: dict[str, Any]) -> None:
        _logger.debug("updating repo webhook '%d' for repo '%s/%s'", webhook_id, org_id, repo_name)

        try:
            await self.requester.request_json("PATCH", f"/repos/{org_id}/{repo_name}/hooks/{webhook_id}", webhook)
            _logger.debug("updated repo webhook '%d'", webhook_id)
        except GitHubException as ex:
            raise RuntimeError(f"failed to update repo webhook {webhook_id}:\n{ex}") from ex

    async def add_webhook(self, org_id: str, repo_name: str, data: dict[str, Any]) -> None:
        url = data["config"]["url"]
        _logger.debug("adding repo webhook with url '%s' for repo '%s/%s'", url, org_id, repo_name)

        # mandatory field "name" = "web"
        data["name"] = "web"

        try:
            await self.requester.request_json("POST", f"/repos/{org_id}/{repo_name}/hooks", data)
            _logger.debug("added repo webhook with url '%s'", url)
        except GitHubException as ex:
            raise RuntimeError(f"failed to add repo webhook with url '{url}':\n{ex}") from ex

    async def delete_webhook(self, org_id: str, repo_name: str, webhook_id: int, url: str) -> None:
        _logger.debug("deleting repo webhook with url '%s' for repo '%s/%s'", url, org_id, repo_name)

        status, _ = await self.requester.request_raw("DELETE", f"/repos/{org_id}/{repo_name}/hooks/{webhook_id}")

        if status != 204:
            raise RuntimeError(f"failed to delete repo webhook with url '{url}'")

        _logger.debug("removed repo webhook with url '%s'", url)

    async def get_ruleset_id(self, org_id: str, repo_name: str, name: str) -> str:
        _logger.debug("retrieving id for repo ruleset with name '%s' for repo '%s/%s'", name, org_id, repo_name)

        rulesets = await self.get_rulesets(org_id, repo_name)

        for ruleset in rulesets:
            if ruleset["name"] == name:
                return ruleset["id"]

        raise RuntimeError(f"failed to find repo ruleset with name '{name}'")

    async def get_rulesets(self, org_id: str, repo_name: str) -> list[dict[str, Any]]:
        _logger.debug("retrieving rulesets for repo '%s/%s'", org_id, repo_name)

        try:
            result = []
            params = {"includes_parents": "false"}
            response = await self.requester.request_paged_json(
                "GET", f"/repos/{org_id}/{repo_name}/rulesets", params=params
            )
            for ruleset in response:
                result.append(await self.get_ruleset(org_id, repo_name, str(ruleset["id"])))
            return result
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving rulesets for repo '{org_id}/{repo_name}':\n{ex}") from ex

    async def get_ruleset(self, org_id: str, repo_name: str, ruleset_id: str) -> dict[str, Any]:
        _logger.debug("retrieving ruleset '%s' for repo '%s/%s'", ruleset_id, org_id, repo_name)

        try:
            params = {"includes_parents": "false"}
            return await self.requester.request_json(
                "GET", f"/repos/{org_id}/{repo_name}/rulesets/{ruleset_id}", params=params
            )
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving ruleset for repo '{org_id}/{repo_name}':\n{ex}") from ex

    async def update_ruleset(self, org_id: str, repo_name: str, ruleset_id: int, ruleset: dict[str, Any]) -> None:
        _logger.debug("updating repo ruleset '%d' for repo '%s/%s'", ruleset_id, org_id, repo_name)

        try:
            await self.requester.request_json("PUT", f"/repos/{org_id}/{repo_name}/rulesets/{ruleset_id}", ruleset)
            _logger.debug("updated repo ruleset '%d'", ruleset_id)
        except GitHubException as ex:
            raise RuntimeError(f"failed to update repo ruleset {ruleset_id}:\n{ex}") from ex

    async def add_ruleset(self, org_id: str, repo_name: str, data: dict[str, Any]) -> None:
        name = data["name"]
        _logger.debug("adding repo ruleset with name '%s' for repo '%s/%s'", name, org_id, repo_name)

        try:
            await self.requester.request_json("POST", f"/repos/{org_id}/{repo_name}/rulesets", data)
            _logger.debug("added repo ruleset with name '%s'", name)
        except GitHubException as ex:
            raise RuntimeError(f"failed to add repo ruleset with name '{name}':\n{ex}") from ex

    async def delete_ruleset(self, org_id: str, repo_name: str, ruleset_id: int, name: str) -> None:
        _logger.debug("deleting repo ruleset with name '%s' for repo '%s/%s'", name, org_id, repo_name)

        status, _ = await self.requester.request_raw("DELETE", f"/repos/{org_id}/{repo_name}/rulesets/{ruleset_id}")

        if status != 204:
            raise RuntimeError(f"failed to delete repo ruleset with name '{name}'")

        _logger.debug("removed repo ruleset with name '%s'", name)

    @staticmethod
    def _render_template_content(org_id: str, repo_name: str, content: str) -> str:
        variables = {"org": org_id, "repo": repo_name}
        return chevron.render(content, variables)

    async def get_readme(self, org_id: str, repo_name: str) -> dict[str, Any]:
        _logger.debug("getting readme for repo '%s/%s'", org_id, repo_name)

        try:
            return await self.requester.request_json("GET", f"/repos/{org_id}/{repo_name}/readme")
        except GitHubException as ex:
            raise RuntimeError(f"failed to get readme for repo '{org_id}/{repo_name}':\n{ex}") from ex

    async def delete_repo(self, org_id: str, repo_name: str) -> None:
        _logger.debug("deleting repo '%s/%s'", org_id, repo_name)

        status, body = await self.requester.request_raw("DELETE", f"/repos/{org_id}/{repo_name}")

        if status != 204:
            raise RuntimeError(f"failed to delete repo '{org_id}/{repo_name}': {body}")

        _logger.debug("removed repo '%s/%s'", org_id, repo_name)

    @staticmethod
    def _remove_already_active_settings(update_data: dict[str, Any], current_data: dict[str, Any]) -> None:
        keys = list(update_data.keys())
        for key in keys:
            if key in current_data:
                update_value_expected = update_data[key]
                update_value_current = current_data[key]

                if update_value_current == update_value_expected:
                    _logger.debug("omitting setting '%s' as it is already set", key)
                    update_data.pop(key)

    async def _fill_github_pages_config(self, org_id: str, repo_name: str, repo_data: dict[str, Any]) -> None:
        _logger.debug("retrieving github pages config for '%s/%s'", org_id, repo_name)

        status, body = await self.requester.request_raw("GET", f"/repos/{org_id}/{repo_name}/pages")
        if status == 200:
            if is_trace_enabled():
                _logger.trace(
                    "'%s' url = %s, json = %s",
                    "GET",
                    f"/repos/{org_id}/{repo_name}/pages",
                    json.dumps(json.loads(body), indent=2),
                )
            repo_data["gh_pages"] = json.loads(body)

    async def _update_github_pages_config(self, org_id: str, repo_name: str, gh_pages: dict[str, Any]) -> None:
        _logger.debug("updating github pages config for '%s/%s'", org_id, repo_name)

        # special handling for repos hosting the organization site
        if repo_name.lower() == f"{org_id}.github.io".lower():
            current_repo_data: dict[str, Any] = {}
            for i in range(1, 4):
                await self._fill_github_pages_config(org_id, repo_name, current_repo_data)
                if "gh_pages" in current_repo_data:
                    break

                _logger.trace(f"waiting for repo '{org_id}/{repo_name}' to be initialized, try {i} of 3")
                import time

                time.sleep(1)

            current_gh_pages: Any = current_repo_data.get("gh_pages")
            if current_gh_pages is not None:
                has_changes = False
                for k, v in gh_pages.items():
                    if current_gh_pages.get(k, None) != v:
                        has_changes = True
                        break

                # if there are no changes to the current config, we do not need to do anything
                if has_changes is False:
                    _logger.trace(f"github pages config for '{org_id}/{repo_name}' is already up-to-date")
                    return

        build_type = gh_pages.get("build_type")
        if build_type == "disabled":
            status, body = await self.requester.request_raw("DELETE", f"/repos/{org_id}/{repo_name}/pages")
            if status != 204 and status != 404:
                raise RuntimeError(f"failed to disable github pages for repo '{repo_name}': {body}")
        else:
            gh_pages_data: list[tuple[str, str, int]] = []
            # first check if the pages config already exists:
            status, _ = await self.requester.request_raw("GET", f"/repos/{org_id}/{repo_name}/pages")
            if status != 200:
                # check if the branch already exists
                source: Any = gh_pages.get("source")
                if source is not None:
                    branch = source.get("branch", None)
                    if branch is not None:
                        existing_branches = await self.get_branches(org_id, repo_name)

                        if len(existing_branches) == 0:
                            _logger.debug("repo '%s' not yet initialized, skipping GH pages config", repo_name)
                            return

                        existing_branch_names = [x["name"] for x in existing_branches]
                        if branch not in existing_branch_names:
                            gh_pages_data.append((json.dumps(gh_pages), "PUT", 204))
                            gh_pages["source"]["branch"] = existing_branch_names[0]
                            gh_pages_data.insert(0, (json.dumps(gh_pages), "POST", 201))

                if len(gh_pages_data) == 0:
                    gh_pages_data.append((json.dumps(gh_pages), "POST", 201))
            else:
                gh_pages_data.append((json.dumps(gh_pages), "PUT", 204))

            for data, method, status_code in gh_pages_data:
                status, body = await self.requester.request_raw(method, f"/repos/{org_id}/{repo_name}/pages", data=data)

                if status != status_code:
                    raise RuntimeError(f"failed to update github pages config for repo '{repo_name}': {body}")

                _logger.debug("updated github pages config for repo '%s'", repo_name)

    async def _fill_code_scanning_config(self, org_id: str, repo_name: str, repo_data: dict[str, Any]) -> None:
        _logger.debug("retrieving code scanning config for '%s/%s'", org_id, repo_name)

        status, body = await self.requester.request_raw(
            "GET", f"/repos/{org_id}/{repo_name}/code-scanning/default-setup"
        )
        if status == 200:
            repo_data["code_scanning_default_config"] = json.loads(body)

    async def _update_code_scanning_config(self, org_id: str, repo_name: str, code_scanning: dict[str, Any]) -> None:
        _logger.debug("updating code scanning config for '%s/%s'", org_id, repo_name)

        try:
            await self.requester.request_json(
                "PATCH", f"/repos/{org_id}/{repo_name}/code-scanning/default-setup", data=code_scanning
            )
            _logger.debug("updated code scanning config for repo '%s/%s'", org_id, repo_name)
        except GitHubException as ex:
            raise RuntimeError(f"failed to update code scanning config for repo '{org_id}/{repo_name}':\n{ex}") from ex

    async def _update_default_branch(self, org_id: str, repo_name: str, new_default_branch: str) -> None:
        _logger.debug("updating default branch for '%s/%s'", org_id, repo_name)
        existing_branches = await self.get_branches(org_id, repo_name)
        existing_branch_names = [x["name"] for x in existing_branches]

        if len(existing_branches) == 0:
            _logger.debug("skip updating of default branch for empty repo '%s/%s'", org_id, repo_name)
            return

        try:
            if new_default_branch in existing_branch_names:
                data = {"default_branch": new_default_branch}
                await self.requester.request_json("PATCH", f"/repos/{org_id}/{repo_name}", data)
                _logger.debug("updated default branch in repo '%s/%s'", org_id, repo_name)
            else:
                default_branch = await self.get_default_branch(org_id, repo_name)
                await self.rename_branch(org_id, repo_name, default_branch, new_default_branch)
                _logger.debug("renamed default branch in repo '%s/%s'", org_id, repo_name)
        except GitHubException as ex:
            raise RuntimeError(f"failed to update default branch for repo '{org_id}/{repo_name}':\n{ex}") from ex

    async def _fill_vulnerability_alerts(self, org_id: str, repo_name: str, repo_data: dict[str, Any]) -> None:
        _logger.debug("retrieving repo vulnerability alerts for '%s/%s'", org_id, repo_name)

        status, _ = await self.requester.request_raw("GET", f"/repos/{org_id}/{repo_name}/vulnerability-alerts")
        if status == 204:
            repo_data["dependabot_alerts_enabled"] = True
        else:
            repo_data["dependabot_alerts_enabled"] = False

    async def _update_vulnerability_alerts(self, org_id: str, repo_name: str, vulnerability_reports: bool) -> None:
        _logger.debug("updating repo vulnerability alerts for '%s/%s'", org_id, repo_name)

        method = "PUT" if vulnerability_reports is True else "DELETE"

        status, body = await self.requester.request_raw(method, f"/repos/{org_id}/{repo_name}/vulnerability-alerts")

        if status != 204:
            raise RuntimeError(f"failed to update vulnerability alerts for repo '{org_id}/{repo_name}': {body}")

        _logger.debug("updated vulnerability alerts for repo '%s/%s'", org_id, repo_name)

    async def _fill_private_vulnerability_reporting(
        self, org_id: str, repo_name: str, repo_data: dict[str, Any]
    ) -> None:
        _logger.debug("retrieving repo private vulnerability reporting status for '%s/%s'", org_id, repo_name)

        response = await self.requester.request_json(
            "GET", f"/repos/{org_id}/{repo_name}/private-vulnerability-reporting"
        )
        repo_data["private_vulnerability_reporting_enabled"] = response["enabled"]

    async def _update_private_vulnerability_reporting(
        self, org_id: str, repo_name: str, private_vulnerability_reporting_enabled: bool
    ) -> None:
        _logger.debug("updating repo private vulnerability reporting for '%s/%s'", org_id, repo_name)

        method = "PUT" if private_vulnerability_reporting_enabled is True else "DELETE"

        status, body = await self.requester.request_raw(
            method, f"/repos/{org_id}/{repo_name}/private-vulnerability-reporting"
        )

        if status != 204:
            raise RuntimeError(
                f"failed to update private vulnerability reporting for repo '{org_id}/{repo_name}': {body}"
            )

        _logger.debug("updated private vulnerability reporting for repo '%s/%s'", org_id, repo_name)

    async def _fill_topics(self, org_id: str, repo_name: str, repo_data: dict[str, Any]) -> None:
        _logger.debug("retrieving repo topics for '%s/%s'", org_id, repo_name)

        try:
            response = await self.requester.request_json("GET", f"/repos/{org_id}/{repo_name}/topics")
            repo_data["topics"] = response.get("names", [])
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving topics for repo '{org_id}/{repo_name}':\n{ex}") from ex

    async def _update_topics(self, org_id: str, repo_name: str, topics: list[str]) -> None:
        _logger.debug("updating repo topics for '%s/%s'", org_id, repo_name)
        data = {"names": topics}
        await self.requester.request_json("PUT", f"/repos/{org_id}/{repo_name}/topics", data=data)
        _logger.debug("updated topics for repo '%s'", repo_name)

    async def _fill_custom_properties(self, org_id: str, repo_name: str, repo_data: dict[str, Any]) -> None:
        _logger.debug("retrieving repo custom properties for '%s/%s'", org_id, repo_name)

        try:
            response = await self.requester.request_json("GET", f"/repos/{org_id}/{repo_name}/properties/values")
            repo_data["custom_properties"] = response
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving custom properties for repo '{org_id}/{repo_name}':\n{ex}") from ex

    async def _update_custom_properties(
        self, org_id: str, repo_name: str, custom_properties: list[dict[str, str | list[str]]]
    ) -> None:
        _logger.debug("updating repo custom properties for '%s/%s'", org_id, repo_name)
        data = {"properties": custom_properties}
        status, body = await self.requester.request_raw(
            "PATCH", f"/repos/{org_id}/{repo_name}/properties/values", data=json.dumps(data)
        )

        if status != 204:
            raise RuntimeError(f"failed to update custom properties for repo '{org_id}/{repo_name}': {body}")

        _logger.debug("updated custom properties for repo '%s'", repo_name)

    async def get_branches(self, org_id: str, repo_name) -> list[dict[str, Any]]:
        _logger.debug("retrieving branches for repo '%s/%s'", org_id, repo_name)

        try:
            return await self.requester.request_paged_json("GET", f"/repos/{org_id}/{repo_name}/branches")
        except GitHubException as ex:
            raise RuntimeError(f"failed getting branches for repo '{org_id}/{repo_name}':\n{ex}") from ex

    async def get_tags(self, org_id: str, repo_name: str) -> list[dict[str, Any]] | None:
        _logger.debug("retrieving tags for repo '%s/%s'", org_id, repo_name)

        try:
            return await self.requester.request_paged_json("GET", f"/repos/{org_id}/{repo_name}/tags")
        except GitHubException as ex:
            raise RuntimeError(f"failed getting tags for repo '{org_id}/{repo_name}':\n{ex}") from ex

    async def get_environments(self, org_id: str, repo_name: str) -> list[dict[str, Any]]:
        _logger.debug("retrieving environments for repo '%s/%s'", org_id, repo_name)

        try:
            response = await self.requester.request_json("GET", f"/repos/{org_id}/{repo_name}/environments")

            environments = response["environments"]
            for env in environments:
                env_name = env["name"]
                has_branch_policies = bool(query_json("deployment_branch_policy.custom_branch_policies", env) or False)

                if has_branch_policies:
                    env["branch_policies"] = await self._get_deployment_branch_policies(org_id, repo_name, env_name)
            return environments
        except GitHubException:
            # querying the environments might fail for private repos, ignore exceptions
            return []

    async def update_environment(self, org_id: str, repo_name: str, env_name: str, env: dict[str, Any]) -> None:
        _logger.debug("updating environment '%s' for repo '%s/%s'", env_name, org_id, repo_name)

        if "name" in env:
            env.pop("name")

        branch_policies = env.pop("branch_policies") if "branch_policies" in env else None

        try:
            await self.requester.request_json("PUT", f"/repos/{org_id}/{repo_name}/environments/{env_name}", env)

            if branch_policies is not None:
                await self._update_deployment_branch_policies(org_id, repo_name, env_name, branch_policies)

            _logger.debug("updated repo environment '%s'", env_name)
        except GitHubException as ex:
            raise RuntimeError(f"failed to update repo environment '{env_name}':\n{ex}") from ex

    async def add_environment(self, org_id: str, repo_name: str, env_name: str, data: dict[str, Any]) -> None:
        _logger.debug("adding environment '%s' for repo '%s/%s'", env_name, org_id, repo_name)
        await self.update_environment(org_id, repo_name, env_name, data)
        _logger.debug("added environment '%s'", env_name)

    async def delete_environment(self, org_id: str, repo_name: str, env_name: str) -> None:
        _logger.debug("deleting repo environment '%s' for repo '%s/%s'", env_name, org_id, repo_name)

        status, _ = await self.requester.request_raw("DELETE", f"/repos/{org_id}/{repo_name}/environments/{env_name}")

        if status != 204:
            raise RuntimeError(f"failed to delete repo environment '{env_name}'")

        _logger.debug("removed repo environment '%s'", env_name)

    async def _get_deployment_branch_policies(self, org_id: str, repo_name: str, env_name: str) -> list[dict[str, Any]]:
        _logger.debug("retrieving deployment branch policies for env '%s'", env_name)

        try:
            url = f"/repos/{org_id}/{repo_name}/environments/{env_name}/deployment-branch-policies"
            response = await self.requester.request_json("GET", url)
            return response["branch_policies"]
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving deployment branch policies:\n{ex}") from ex

    async def _update_deployment_branch_policies(
        self, org_id: str, repo_name: str, env_name: str, branch_policies: list[str]
    ) -> None:
        _logger.debug("updating deployment branch policies for env '%s'", env_name)

        try:
            current_branch_policies_by_name = associate_by_key(
                await self._get_deployment_branch_policies(org_id, repo_name, env_name),
                lambda x: f"{x.get('type', 'branch')}:{x['name']}",
            )
        except RuntimeError:
            current_branch_policies_by_name = {}

        try:

            def transform_policies(x: str):
                if x.startswith("tag:"):
                    return f"tag:{x[4:]}"
                else:
                    return f"branch:{x}"

            for policy in map(transform_policies, branch_policies):
                if policy in current_branch_policies_by_name:
                    current_branch_policies_by_name.pop(policy)
                else:
                    await self._create_deployment_branch_policy(org_id, repo_name, env_name, policy)

            for _policy_name, policy_dict in current_branch_policies_by_name.items():
                await self._delete_deployment_branch_policy(org_id, repo_name, env_name, policy_dict["id"])

            _logger.debug("updated deployment branch policies for env '%s'", env_name)

        except GitHubException as ex:
            raise RuntimeError(f"failed creating deployment branch policies:\n{ex}") from ex

    async def _create_deployment_branch_policy(
        self, org_id: str, repo_name: str, env_name: str, name_and_type: str
    ) -> None:
        _logger.debug("creating deployment branch policy for env '%s' with type/name '%s", env_name, name_and_type)

        try:
            target_type, name = name_and_type.split(":", 1)
            data = {"name": name, "type": target_type}
            url = f"/repos/{org_id}/{repo_name}/environments/{env_name}/deployment-branch-policies"
            await self.requester.request_json("POST", url, data)
            _logger.debug("created deployment branch policy for env '%s'", env_name)
        except GitHubException as ex:
            raise RuntimeError(f"failed creating deployment branch policy:\n{ex}") from ex

    async def _delete_deployment_branch_policy(
        self, org_id: str, repo_name: str, env_name: str, policy_id: int
    ) -> None:
        _logger.debug("deleting deployment branch policy for env '%s' with id '%d", env_name, policy_id)

        url = f"/repos/{org_id}/{repo_name}/environments/{env_name}/deployment-branch-policies/{policy_id}"
        status, body = await self.requester.request_raw("DELETE", url)

        if status != 204:
            raise RuntimeError(f"failed deleting deployment branch policy\n{status}: {body}")

        _logger.debug("deleted deployment branch policy for env '%s'", env_name)

    async def get_secrets(self, org_id: str, repo_name: str) -> list[dict[str, Any]]:
        _logger.debug("retrieving secrets for repo '%s/%s'", org_id, repo_name)

        try:
            status, body = await self.requester.request_raw("GET", f"/repos/{org_id}/{repo_name}/actions/secrets")
            if status == 200:
                return json.loads(body)["secrets"]
            else:
                return []
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving secrets for repo '{org_id}/{repo_name}':\n{ex}") from ex

    async def update_secret(self, org_id: str, repo_name: str, secret_name: str, secret: dict[str, Any]) -> None:
        _logger.debug("updating repo secret '%s' for repo '%s/%s'", secret_name, org_id, repo_name)

        if "name" in secret:
            secret.pop("name")

        await self._encrypt_secret_inplace(org_id, repo_name, secret)

        status, _ = await self.requester.request_raw(
            "PUT",
            f"/repos/{org_id}/{repo_name}/actions/secrets/{secret_name}",
            json.dumps(secret),
        )

        if status != 204:
            raise RuntimeError(f"failed to update repo secret '{secret_name}'")

        _logger.debug("updated repo secret '%s'", secret_name)

    async def add_secret(self, org_id: str, repo_name: str, data: dict[str, str]) -> None:
        secret_name = data.pop("name")
        _logger.debug("adding repo secret '%s' for repo '%s/%s'", secret_name, org_id, repo_name)

        await self._encrypt_secret_inplace(org_id, repo_name, data)

        status, _ = await self.requester.request_raw(
            "PUT",
            f"/repos/{org_id}/{repo_name}/actions/secrets/{secret_name}",
            json.dumps(data),
        )

        if status != 201:
            raise RuntimeError(f"failed to add repo secret '{secret_name}'")

        _logger.debug("added repo secret '%s'", secret_name)

    async def _encrypt_secret_inplace(self, org_id: str, repo_name: str, data: dict[str, Any]) -> None:
        value = data.pop("value")
        key_id, public_key = await self.get_public_key(org_id, repo_name)
        data["encrypted_value"] = encrypt_value(public_key, value)
        data["key_id"] = key_id

    async def delete_secret(self, org_id: str, repo_name: str, secret_name: str) -> None:
        _logger.debug("deleting repo secret '%s' for repo '%s/%s'", secret_name, org_id, repo_name)

        status, _ = await self.requester.request_raw(
            "DELETE", f"/repos/{org_id}/{repo_name}/actions/secrets/{secret_name}"
        )

        if status != 204:
            raise RuntimeError(f"failed to delete repo secret '{secret_name}'")

        _logger.debug("removed repo secret '%s'", secret_name)

    async def get_variables(self, org_id: str, repo_name: str) -> list[dict[str, Any]]:
        _logger.debug("retrieving variables for repo '%s/%s'", org_id, repo_name)

        try:
            status, body = await self.requester.request_raw("GET", f"/repos/{org_id}/{repo_name}/actions/variables")
            if status == 200:
                return json.loads(body)["variables"]
            else:
                return []
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving variables for repo '{org_id}/{repo_name}':\n{ex}") from ex

    async def update_variable(self, org_id: str, repo_name: str, variable_name: str, variable: dict[str, Any]) -> None:
        _logger.debug("updating repo variable '%s' for repo '%s/%s'", variable_name, org_id, repo_name)

        if "name" in variable:
            variable.pop("name")

        status, body = await self.requester.request_raw(
            "PATCH",
            f"/repos/{org_id}/{repo_name}/actions/variables/{variable_name}",
            json.dumps(variable),
        )
        if status != 204:
            raise RuntimeError(f"failed to update repo variable '{variable_name}': {body}")

        _logger.debug("updated repo variable '%s'", variable_name)

    async def add_variable(self, org_id: str, repo_name: str, data: dict[str, str]) -> None:
        variable_name = data.get("name")
        _logger.debug("adding repo variable '%s' for repo '%s/%s'", variable_name, org_id, repo_name)

        status, body = await self.requester.request_raw(
            "POST",
            f"/repos/{org_id}/{repo_name}/actions/variables",
            json.dumps(data),
        )

        if status != 201:
            raise RuntimeError(f"failed to add repo variable '{variable_name}': {body}")

        _logger.debug("added repo variable '%s'", variable_name)

    async def delete_variable(self, org_id: str, repo_name: str, variable_name: str) -> None:
        _logger.debug("deleting repo variable '%s' for repo '%s/%s'", variable_name, org_id, repo_name)

        status, _ = await self.requester.request_raw(
            "DELETE", f"/repos/{org_id}/{repo_name}/actions/variables/{variable_name}"
        )

        if status != 204:
            raise RuntimeError(f"failed to delete repo variable '{variable_name}'")

        _logger.debug("removed repo variable '%s'", variable_name)

    async def get_workflow_settings(self, org_id: str, repo_name: str) -> dict[str, Any]:
        _logger.debug("retrieving workflow settings for repo '%s/%s'", org_id, repo_name)

        workflow_settings: dict[str, Any] = {}

        try:
            permissions = await self.requester.request_json("GET", f"/repos/{org_id}/{repo_name}/actions/permissions")
            workflow_settings.update(permissions)
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving workflow settings for repo '{org_id}/{repo_name}':\n{ex}") from ex

        allowed_actions = permissions.get("allowed_actions", "none")
        if allowed_actions == "selected":
            workflow_settings.update(await self._get_selected_actions_for_workflow_settings(org_id, repo_name))

        if permissions.get("enabled", False) is not False:
            workflow_settings.update(await self._get_default_workflow_permissions(org_id, repo_name))

        return workflow_settings

    async def update_workflow_settings(self, org_id: str, repo_name: str, data: dict[str, Any]) -> None:
        _logger.debug("updating workflow settings for repo '%s/%s'", org_id, repo_name)

        permission_data = {k: data[k] for k in ["enabled", "allowed_actions"] if k in data}
        if len(permission_data) > 0:
            status, body = await self.requester.request_raw(
                "PUT", f"/repos/{org_id}/{repo_name}/actions/permissions", json.dumps(permission_data)
            )

            if status != 204:
                raise RuntimeError(
                    f"failed to update workflow settings for repo '{org_id}/{repo_name}'\n{status}: {body}"
                )

            _logger.debug("updated workflow settings for repo '%s/%s'", org_id, repo_name)

        # only update the selected actions if needed
        if data.get("allowed_actions", "selected") == "selected":
            allowed_action_data = {
                k: data[k] for k in ["github_owned_allowed", "verified_allowed", "patterns_allowed"] if k in data
            }
            if len(allowed_action_data) > 0:
                await self._update_selected_actions_for_workflow_settings(org_id, repo_name, allowed_action_data)

        default_permission_data = {
            k: data[k] for k in ["default_workflow_permissions", "can_approve_pull_request_reviews"] if k in data
        }
        if len(default_permission_data) > 0:
            await self._update_default_workflow_permissions(org_id, repo_name, default_permission_data)

        _logger.debug("updated %d workflow setting(s)", len(data))

    async def _get_selected_actions_for_workflow_settings(self, org_id: str, repo_name: str) -> dict[str, Any]:
        _logger.debug("retrieving allowed actions for repo '%s/%s'", org_id, repo_name)

        try:
            return await self.requester.request_json(
                "GET", f"/repos/{org_id}/{repo_name}/actions/permissions/selected-actions"
            )
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving allowed actions for repo '{org_id}/{repo_name}':\n{ex}") from ex

    async def _update_selected_actions_for_workflow_settings(
        self, org_id: str, repo_name: str, data: dict[str, Any]
    ) -> None:
        _logger.debug("updating allowed actions for repo '%s/%s'", org_id, repo_name)

        status, body = await self.requester.request_raw(
            "PUT", f"/repos/{org_id}/{repo_name}/actions/permissions/selected-actions", json.dumps(data)
        )

        if status != 204:
            raise RuntimeError(f"failed updating allowed actions for repo '{org_id}/{repo_name}'\n{status}: {body}")

        _logger.debug("updated allowed actions for repo '%s/%s'", org_id, repo_name)

    async def _get_default_workflow_permissions(self, org_id: str, repo_name: str) -> dict[str, Any]:
        _logger.debug("retrieving default workflow permissions for repo '%s/%s'", org_id, repo_name)

        try:
            return await self.requester.request_json("GET", f"/repos/{org_id}/{repo_name}/actions/permissions/workflow")
        except GitHubException as ex:
            raise RuntimeError(
                f"failed retrieving default workflow permissions for repo '{org_id}/{repo_name}':\n{ex}"
            ) from ex

    async def _update_default_workflow_permissions(self, org_id: str, repo_name: str, data: dict[str, Any]) -> None:
        _logger.debug("updating default workflow permissions for repo '%s/%s'", org_id, repo_name)

        status, body = await self.requester.request_raw(
            "PUT", f"/repos/{org_id}/{repo_name}/actions/permissions/workflow", json.dumps(data)
        )

        if status != 204:
            raise RuntimeError(
                f"failed updating default workflow permissions for repo '{org_id}/{repo_name}'\n{status}: {body}"
            )

        _logger.debug("updated default workflow permissions for repo '%s/%s'", org_id, repo_name)

    async def get_public_key(self, org_id: str, repo_name: str) -> tuple[str, str]:
        _logger.debug("retrieving repo public key for repo '%s/%s'", org_id, repo_name)

        try:
            response = await self.requester.request_json(
                "GET", f"/repos/{org_id}/{repo_name}/actions/secrets/public-key"
            )
            return response["key_id"], response["key"]
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving repo public key:\n{ex}") from ex

    async def dispatch_workflow(self, org_id: str, repo_name: str, workflow_name: str) -> bool:
        _logger.debug("dispatching workflow for repo '%s/%s'", org_id, repo_name)

        default_branch = await self.get_default_branch(org_id, repo_name)
        data = {"ref": default_branch}

        status, _ = await self.requester.request_raw(
            "POST", f"/repos/{org_id}/{repo_name}/actions/workflows/{workflow_name}/dispatches", json.dumps(data)
        )

        if status != 204:
            _logger.debug("failed dispatching workflow for repo '%s/%s'", org_id, repo_name)
            return False
        else:
            _logger.debug("dispatched workflow for repo '%s/%s'", org_id, repo_name)
            return True

    async def get_ref_for_pull_request(self, org_id: str, repo_name: str, pull_number: str) -> str:
        _logger.debug(f"retrieving ref for pull request {pull_number} at {org_id}/{repo_name}")

        try:
            response = await self.requester.request_json("GET", f"/repos/{org_id}/{repo_name}/pulls/{pull_number}")
            return response["head"]["sha"]
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving ref for pull request:\n{ex}") from ex

    async def sync_from_template_repository(
        self,
        org_id: str,
        repo_name: str,
        template_repository: str,
        template_paths: list[str] | None,
    ) -> list[str]:
        template_owner, template_repo = re.split("/", template_repository, maxsplit=1)

        updated_files = []
        with tempfile.TemporaryDirectory() as tmp_dir:
            archive_file_name = os.path.join(tmp_dir, "archive.zip")
            async with aiofiles.open(archive_file_name, "wb") as archive_file:
                await self._download_repository_archive(archive_file, template_owner, template_repo)

            archive_target_dir = os.path.join(tmp_dir, "contents")
            with zipfile.ZipFile(archive_file_name, "r") as zip_file:
                zip_file.extractall(archive_target_dir)

            template_paths_set = set(template_paths) if isinstance(template_paths, list) else set()

            base_dir = None
            for path in pathlib.Path(archive_target_dir).rglob("*"):
                # the downloaded archive starts with a subdir that encodes
                # the name / hash of the downloaded repo, use that as the base dir
                # to resolve relative path names for updating the content.
                if base_dir is None:
                    base_dir = path

                relative_path = path.relative_to(base_dir)

                if path.is_file():
                    _logger.debug("updating file '%s'", relative_path)

                    with open(path) as file:
                        content = file.read()

                        if str(relative_path) in template_paths_set:
                            content = self._render_template_content(org_id, repo_name, content)

                        updated = await self.rest_api.content.update_content(
                            org_id, repo_name, str(relative_path), content
                        )
                        if updated:
                            updated_files.append(str(relative_path))

        return updated_files

    async def _download_repository_archive(self, file, org_id: str, repo_name: str, ref: str = "") -> None:
        _logger.debug("downloading repository archive for '%s/%s'", org_id, repo_name)

        try:
            async for data in self.requester.request_stream("GET", f"/repos/{org_id}/{repo_name}/zipball/{ref}"):
                await file.write(data)

        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving repository archive from repo '{org_id}/{repo_name}':\n{ex}") from ex



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

from typing import TYPE_CHECKING

from otterdog.providers.github import GitHubProvider
from otterdog.utils import format_date_for_csv, is_info_enabled

from . import Operation

if TYPE_CHECKING:
    from otterdog.config import OrganizationConfig


class ListAdvisoriesOperation(Operation):
    """
    Lists repository security advisories for an organization.
    """

    def __init__(self, states: list[str], details: bool):
        super().__init__()
        # if states contains "all", then we will get all advisories
        if "all" in states:
            states = ["triage", "draft", "published", "closed"]
        self._states = states
        self._details = details

    @property
    def states(self) -> list[str]:
        return self._states

    @property
    def details(self) -> bool:
        return self._details

    def pre_execute(self) -> None:
        if is_info_enabled():
            self.printer.println(f"Listing {self.states} repository security advisories:")
        if not self.details:
            self.printer.println(
                "organization,created_at,updated_at,published_at,state,severity,ghsa_id,cve_id,html_url,summary"
            )

    def post_execute(self) -> None:
        pass

    async def execute(
        self,
        org_config: OrganizationConfig,
        org_index: int | None = None,
        org_count: int | None = None,
    ) -> int:
        github_id = org_config.github_id

        if is_info_enabled():
            self._print_project_header(org_config, org_index, org_count)
            self.printer.level_up()

        try:
            try:
                credentials = self.get_credentials(org_config, only_token=True)
            except RuntimeError as e:
                self.printer.print_error(f"invalid credentials\n{e!s}")
                return 1

            advisories = []
            for state in self.states:
                async with GitHubProvider(credentials) as provider:
                    advisories_for_state = await provider.rest_api.org.get_security_advisories(github_id, state)
                    advisories += advisories_for_state

            if not self.details:
                if is_info_enabled():
                    self.printer.println(f"Found {len(advisories)} advisories with state '{self.states}'.")
                    self.printer.println()

            for advisory in advisories:
                if not self.details:
                    cve_id = advisory["cve_id"] if advisory["cve_id"] is not None else "NO_CVE"
                    summary = advisory["summary"].replace('"', '""')

                    formatted_values = {
                        "org_name": org_config.name,
                        "created_at": format_date_for_csv(advisory["created_at"]),
                        "updated_at": format_date_for_csv(advisory["updated_at"]),
                        "published_at": format_date_for_csv(advisory["published_at"]),
                        "state": advisory["state"],
                        "severity": advisory["severity"],
                        "ghsa_id": advisory["ghsa_id"],
                        "cve_id": cve_id,
                        "html_url": advisory["html_url"],
                        "summary": summary,
                    }

                    self.printer.println(
                        f"\"{formatted_values['org_name']}\","
                        f"\"{formatted_values['created_at']}\","
                        f"\"{formatted_values['updated_at']}\","
                        f"\"{formatted_values['published_at']}\","
                        f"\"{formatted_values['state']}\","
                        f"\"{formatted_values['severity']}\","
                        f"\"{formatted_values['ghsa_id']}\","
                        f"\"{formatted_values['cve_id']}\","
                        f"\"{formatted_values['html_url']}\","
                        f"\"{formatted_values['summary']}\""
                    )
                else:
                    self.print_dict(advisory, f"advisory['{advisory['ghsa_id']}']", "", "black")

            return 0
        finally:
            if is_info_enabled():
                self.printer.level_down()



#  *******************************************************************************
#  Copyright (c) 2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

from typing import TYPE_CHECKING

from otterdog.logging import get_logger
from otterdog.providers.github.cache.file import file_cache

if TYPE_CHECKING:
    from otterdog.providers.github.cache import CacheStrategy

_GITHUB_CACHE = file_cache()
_logger = get_logger(__name__)


def get_github_cache() -> CacheStrategy:
    global _GITHUB_CACHE

    _logger.trace("using %s as GitHub cache strategy", _GITHUB_CACHE)
    return _GITHUB_CACHE


def set_github_cache(cache: CacheStrategy) -> None:
    global _GITHUB_CACHE

    _logger.trace("setting %s as GitHub cache strategy", cache)
    _GITHUB_CACHE = cache



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import os
import unittest

from otterdog.config import OtterdogConfig
from otterdog.models.github_organization import GitHubOrganization


class GitHubOrganizationTest(unittest.IsolatedAsyncioTestCase):
    TEST_ORG = "test-org"
    BASE_TEMPLATE_URL = "https://github.com/otterdog/test-defaults#test-defaults.libsonnet@main"

    async def asyncSetUp(self):
        base_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), "resources")
        otterdog_config_file = os.path.join(base_dir, "otterdog.json")

        self.otterdog_config = OtterdogConfig.from_file(otterdog_config_file, True)
        self.org_config = self.otterdog_config.get_organization_config(self.TEST_ORG)
        self.jsonnet_config = self.org_config.jsonnet_config
        await self.jsonnet_config.init_template()

    def test_load_from_file(self):
        organization = GitHubOrganization.load_from_file(self.TEST_ORG, self.jsonnet_config.org_config_file)

        assert organization.github_id == "test-org"
        assert len(organization.webhooks) == 1
        assert len(organization.repositories) == 2



#  *******************************************************************************
#  Copyright (c) 2023-2025 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from collections.abc import Mapping
from typing import Any

from otterdog.jsonnet import JsonnetConfig
from otterdog.models import ModelObject
from otterdog.models.branch_protection_rule import BranchProtectionRule
from otterdog.utils import UNSET, Change

from . import ModelTest


class BranchProtectionRuleTest(ModelTest):
    def create_model(self, data: Mapping[str, Any]) -> ModelObject:
        return BranchProtectionRule.from_model_data(data)

    @property
    def template_function(self) -> str:
        return JsonnetConfig.create_branch_protection_rule

    @property
    def model_data(self):
        return self.load_json_resource("otterdog-bpr.json")

    @property
    def provider_data(self):
        return self.load_json_resource("github-bpr.json")

    def test_load_from_model(self):
        bpr = BranchProtectionRule.from_model_data(self.model_data)

        assert bpr.id is UNSET
        assert bpr.pattern == "main"
        assert bpr.allows_deletions is False
        assert bpr.allows_force_pushes is False
        assert bpr.dismisses_stale_reviews is False
        assert bpr.is_admin_enforced is False
        assert bpr.lock_allows_fetch_and_merge is False
        assert bpr.lock_branch is False
        assert bpr.bypass_force_push_allowances == ["@netomi"]
        assert bpr.bypass_pull_request_allowances == ["@netomi"]
        assert bpr.push_restrictions == ["@netomi"]
        assert bpr.require_last_push_approval is False
        assert bpr.required_approving_review_count == 2
        assert bpr.requires_pull_request is True
        assert bpr.requires_code_owner_reviews is False
        assert bpr.requires_commit_signatures is False
        assert bpr.requires_conversation_resolution is False
        assert bpr.requires_linear_history is False
        assert bpr.requires_status_checks is True
        assert bpr.requires_strict_status_checks is False
        assert bpr.restricts_review_dismissals is False
        assert bpr.review_dismissal_allowances == ["@netomi"]
        assert bpr.required_status_checks == ["eclipse-eca-validation:eclipsefdn/eca", "any:Run CI"]

    def test_load_from_provider(self):
        bpr = BranchProtectionRule.from_provider_data(self.org_id, self.provider_data)

        assert bpr.id == "BPR_kwDOI9xAhM4CC77t"
        assert bpr.pattern == "main"
        assert bpr.allows_deletions is False
        assert bpr.allows_force_pushes is False
        assert bpr.dismisses_stale_reviews is False
        assert bpr.is_admin_enforced is False
        assert bpr.lock_allows_fetch_and_merge is False
        assert bpr.lock_branch is False
        assert bpr.bypass_force_push_allowances == ["@netomi"]
        assert bpr.bypass_pull_request_allowances == ["@netomi"]
        assert bpr.push_restrictions == ["@netomi"]
        assert bpr.require_last_push_approval is False
        assert bpr.required_approving_review_count == 2
        assert bpr.requires_pull_request is True
        assert bpr.requires_code_owner_reviews is False
        assert bpr.requires_commit_signatures is False
        assert bpr.requires_conversation_resolution is False
        assert bpr.requires_linear_history is False
        assert bpr.requires_status_checks is True
        assert bpr.requires_strict_status_checks is False
        assert bpr.restricts_review_dismissals is False
        assert bpr.review_dismissal_allowances == ["@netomi"]
        assert bpr.required_status_checks == ["any:Run CI"]

    async def test_to_provider(self):
        bpr = BranchProtectionRule.from_model_data(self.model_data)

        provider_data = await bpr.to_provider_data(self.org_id, self.provider)

        assert len(provider_data) == 24
        assert provider_data["pattern"] == "main"
        assert provider_data["pushActorIds"] == ["id_netomi"]
        assert provider_data["requiredStatusChecks"] == [
            {"appId": "id_eclipse-eca-validation", "context": "eclipsefdn/eca"},
            {"appId": "any", "context": "Run CI"},
        ]

    async def test_changes_to_provider(self):
        current = BranchProtectionRule.from_model_data(self.model_data)
        other = BranchProtectionRule.from_model_data(self.model_data)

        other.requires_pull_request = False
        other.required_status_checks = ["eclipse-eca-validation:eclipsefdn/eca"]

        changes = current.get_difference_from(other)
        provider_data = await BranchProtectionRule.changes_to_provider(self.org_id, changes, self.provider)

        assert len(provider_data) == 2
        assert provider_data["requiresApprovingReviews"] is True
        assert provider_data["requiredStatusChecks"] == [
            {"appId": "id_eclipse-eca-validation", "context": "eclipsefdn/eca"},
            {"appId": "any", "context": "Run CI"},
        ]

    def test_patch(self):
        current = BranchProtectionRule.from_model_data(self.model_data)
        default = BranchProtectionRule.from_model_data(self.model_data)

        default.pattern = None
        default.requires_status_checks = False
        default.required_status_checks = ["eclipse-eca-validation:eclipsefdn/eca"]

        patch = current.get_patch_to(default)

        assert len(patch) == 3
        assert patch["pattern"] == current.pattern
        assert patch["requires_status_checks"] is current.requires_status_checks
        assert patch["required_status_checks"] == ["any:Run CI"]

    def test_difference(self):
        current = BranchProtectionRule.from_model_data(self.model_data)
        other = BranchProtectionRule.from_model_data(self.model_data)

        other.requires_pull_request = False
        other.required_status_checks = ["eclipse-eca-validation:eclipsefdn/eca"]

        diff = current.get_difference_from(other)

        assert len(diff) == 2
        assert diff["requires_pull_request"] == Change(other.requires_pull_request, current.requires_pull_request)
        assert diff["required_status_checks"] == Change(other.required_status_checks, current.required_status_checks)



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

import dataclasses
import fnmatch
from typing import TYPE_CHECKING, Any, cast

from jsonbender import K, OptionalS  # type: ignore

from otterdog.models import FailureType, LivePatch, LivePatchType, ValidationContext
from otterdog.models.ruleset import Ruleset
from otterdog.utils import is_set_and_valid, unwrap

if TYPE_CHECKING:
    from otterdog.jsonnet import JsonnetConfig
    from otterdog.providers.github import GitHubProvider


@dataclasses.dataclass
class OrganizationRuleset(Ruleset):
    """
    Represents a ruleset defined on organization level.
    """

    include_repo_names: list[str]
    exclude_repo_names: list[str]
    protect_repo_names: bool

    @property
    def model_object_name(self) -> str:
        return "org_ruleset"

    def get_jsonnet_template_function(self, jsonnet_config: JsonnetConfig, extend: bool) -> str | None:
        return f"orgs.{jsonnet_config.create_org_ruleset}"

    def validate(self, context: ValidationContext, parent_object: Any) -> None:
        from otterdog.models.github_organization import GitHubOrganization

        super().validate(context, parent_object)

        repositories = cast(GitHubOrganization, context.root_object).repositories
        all_repo_names = (x.name for x in repositories)

        if is_set_and_valid(self.include_repo_names):
            for repo_name_pattern in self.include_repo_names:
                if repo_name_pattern == "~ALL":
                    continue
                elif len(fnmatch.filter(all_repo_names, repo_name_pattern)) == 0:
                    context.add_failure(
                        FailureType.WARNING,
                        f"{self.get_model_header(parent_object)} has an 'include_repo_names' pattern "
                        f"'{repo_name_pattern}' that does not match any existing repository",
                    )

        if is_set_and_valid(self.exclude_repo_names):
            for repo_name_pattern in self.exclude_repo_names:
                if repo_name_pattern == "~ALL":
                    continue
                elif len(fnmatch.filter(all_repo_names, repo_name_pattern)) == 0:
                    context.add_failure(
                        FailureType.WARNING,
                        f"{self.get_model_header(parent_object)} has an 'exclude_repo_names' pattern "
                        f"'{repo_name_pattern}' that does not match any existing repository",
                    )

        if is_set_and_valid(self.protect_repo_names):
            if (
                self.protect_repo_names is True
                and len(self.include_repo_names) == 0
                and len(self.exclude_repo_names) == 0
            ):
                context.add_failure(
                    FailureType.WARNING,
                    f"{self.get_model_header(parent_object)} has 'protect_repo_names' set to "
                    f"'{self.protect_repo_names}' but 'include_repo_names' and 'exclude_repo_names' are empty.",
                )

    @classmethod
    def get_mapping_from_provider(cls, org_id: str, data: dict[str, Any]) -> dict[str, Any]:
        mapping = super().get_mapping_from_provider(org_id, data)

        mapping.update(
            {
                "include_repo_names": OptionalS("conditions", "repository_name", "include", default=[]),
                "exclude_repo_names": OptionalS("conditions", "repository_name", "exclude", default=[]),
                "protect_repo_names": OptionalS("conditions", "repository_name", "protected", default=False),
            }
        )

        return mapping

    @classmethod
    async def get_mapping_to_provider(
        cls, org_id: str, data: dict[str, Any], provider: GitHubProvider
    ) -> dict[str, Any]:
        mapping = await super().get_mapping_to_provider(org_id, data, provider)

        # include_repo_names / exclude_repo_names
        repo_names = {}
        if "include_repo_names" in data:
            mapping.pop("include_repo_names")
            repo_names["include"] = K(data["include_repo_names"])

        if "exclude_repo_names" in data:
            mapping.pop("exclude_repo_names")
            repo_names["exclude"] = K(data["exclude_repo_names"])

        if "protect_repo_names" in data:
            mapping.pop("protect_repo_names")
            repo_names["protected"] = K(data["protect_repo_names"])

        if len(repo_names) > 0:
            conditions = mapping.get("conditions", {})
            conditions.update({"repository_name": repo_names})

        return mapping

    @classmethod
    async def apply_live_patch(
        cls,
        patch: LivePatch[OrganizationRuleset],
        org_id: str,
        provider: GitHubProvider,
    ) -> None:
        match patch.patch_type:
            case LivePatchType.ADD:
                await provider.add_org_ruleset(
                    org_id,
                    await unwrap(patch.expected_object).to_provider_data(org_id, provider),
                )

            case LivePatchType.REMOVE:
                current_object = unwrap(patch.current_object)
                await provider.delete_org_ruleset(org_id, current_object.id, current_object.name)

            case LivePatchType.CHANGE:
                current_object = unwrap(patch.current_object)
                await provider.update_org_ruleset(
                    org_id,
                    current_object.id,
                    current_object.name,
                    await unwrap(patch.expected_object).to_provider_data(org_id, provider),
                )



#  *******************************************************************************
#  Copyright (c) 2023-2025 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from collections.abc import Mapping
from typing import Any

from otterdog.jsonnet import JsonnetConfig
from otterdog.models import ModelObject
from otterdog.models.organization_webhook import OrganizationWebhook
from otterdog.utils import UNSET, Change, query_json

from . import ModelTest


class OrganizationWebhookTest(ModelTest):
    def create_model(self, data: Mapping[str, Any]) -> ModelObject:
        return OrganizationWebhook.from_model_data(data)

    @property
    def template_function(self) -> str:
        return JsonnetConfig.create_org_webhook

    @property
    def model_data(self):
        return self.load_json_resource("otterdog-webhook.json")

    @property
    def provider_data(self):
        return self.load_json_resource("github-webhook.json")

    def test_load_from_model(self):
        webhook = OrganizationWebhook.from_model_data(self.model_data)

        assert webhook.id is UNSET
        assert webhook.active is True
        assert webhook.events == ["push"]
        assert webhook.content_type == "form"
        assert webhook.secret == "blabla"
        assert webhook.url == "https://www.example.org"
        assert webhook.insecure_ssl == "0"

    def test_load_from_provider(self):
        webhook = OrganizationWebhook.from_provider_data(self.org_id, self.provider_data)

        assert webhook.id == 1
        assert webhook.active is True
        assert webhook.events == ["push", "pull_request"]
        assert webhook.content_type == "json"
        assert webhook.secret is None
        assert webhook.url == "https://www.example.org"
        assert webhook.insecure_ssl == "0"

    async def test_to_provider(self):
        webhook = OrganizationWebhook.from_model_data(self.model_data)

        webhook.secret = UNSET

        provider_data = await webhook.to_provider_data(self.org_id, self.provider)

        assert len(provider_data) == 3
        assert provider_data["active"] is True
        assert provider_data["events"] == ["push"]

        assert query_json("config.secret", provider_data) or "" == ""
        assert query_json("config.url", provider_data) == "https://www.example.org"
        assert query_json("config.insecure_ssl", provider_data) == "0"
        assert query_json("config.content_type", provider_data) == "form"

    async def test_changes_to_provider(self):
        current = OrganizationWebhook.from_model_data(self.model_data)
        other = OrganizationWebhook.from_model_data(self.model_data)

        other.active = False
        other.insecure_ssl = "1"

        changes = current.get_difference_from(other)
        provider_data = await OrganizationWebhook.changes_to_provider(self.org_id, changes, self.provider)

        assert len(provider_data) == 2
        assert provider_data["active"] is True
        assert query_json("config.insecure_ssl", provider_data) == "0"

    def test_patch(self):
        current = OrganizationWebhook.from_model_data(self.model_data)
        default = OrganizationWebhook.from_model_data(self.model_data)

        default.url = "https://www.notexistent.org"
        default.active = False

        patch = current.get_patch_to(default)

        assert len(patch) == 2
        assert patch["url"] == current.url
        assert patch["active"] is current.active

    def test_difference(self):
        current = OrganizationWebhook.from_model_data(self.model_data)
        other = OrganizationWebhook.from_model_data(self.model_data)

        other.active = False
        other.insecure_ssl = "1"

        diff = current.get_difference_from(other)

        assert len(diff) == 2
        assert diff["active"] == Change(other.active, current.active)
        assert diff["insecure_ssl"] == Change(other.insecure_ssl, current.insecure_ssl)



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************


from otterdog.logging import get_logger

from . import RestApi, RestClient

_logger = get_logger(__name__)


class MetaClient(RestClient):
    def __init__(self, rest_api: RestApi):
        super().__init__(rest_api)

    async def get_scopes(self) -> str:
        _logger.debug("retrieving token scopes")

        status, body, scopes = await self.requester.request_raw_with_scopes("GET", "/rate_limit")
        if status == 200:
            return scopes
        else:
            raise RuntimeError(f"failed retrieving token scopes:\n{body}")



#  *******************************************************************************
#  Copyright (c) 2023-2025 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

import asyncio
import dataclasses
import json
import os
from io import StringIO
from typing import TYPE_CHECKING, Any

from importlib_resources import as_file, files
from jsonbender import F, Forall, OptionalS, S, bend  # type: ignore

from otterdog import resources
from otterdog.logging import get_logger
from otterdog.models import (
    FailureType,
    LivePatchContext,
    LivePatchHandler,
    ModelObject,
    PatchContext,
    ValidationContext,
)
from otterdog.models.branch_protection_rule import BranchProtectionRule
from otterdog.models.custom_property import CustomProperty
from otterdog.models.environment import Environment
from otterdog.models.organization_role import OrganizationRole
from otterdog.models.organization_ruleset import OrganizationRuleset
from otterdog.models.organization_secret import OrganizationSecret
from otterdog.models.organization_settings import OrganizationSettings
from otterdog.models.organization_variable import OrganizationVariable
from otterdog.models.organization_webhook import OrganizationWebhook
from otterdog.models.organization_workflow_settings import OrganizationWorkflowSettings
from otterdog.models.repo_ruleset import RepositoryRuleset
from otterdog.models.repo_secret import RepositorySecret
from otterdog.models.repo_variable import RepositoryVariable
from otterdog.models.repo_webhook import RepositoryWebhook
from otterdog.models.repo_workflow_settings import RepositoryWorkflowSettings
from otterdog.models.repository import Repository
from otterdog.models.team import Team
from otterdog.utils import IndentingPrinter, associate_by_key, debug_times, jsonnet_evaluate_file

if TYPE_CHECKING:
    from collections.abc import AsyncIterator, Callable, Iterator
    from re import Pattern

    from otterdog.config import JsonnetConfig, OtterdogConfig, SecretResolver
    from otterdog.providers.github import GitHubProvider

_ORG_SCHEMA = json.loads(files(resources).joinpath("schemas/organization.json").read_text())

_logger = get_logger(__name__)


@dataclasses.dataclass
class GitHubOrganization:
    """
    Represents a GitHub Organization with its associated resources.
    """

    project_name: str
    github_id: str
    settings: OrganizationSettings
    roles: list[OrganizationRole] = dataclasses.field(default_factory=list)
    teams: list[Team] = dataclasses.field(default_factory=list)
    webhooks: list[OrganizationWebhook] = dataclasses.field(default_factory=list)
    secrets: list[OrganizationSecret] = dataclasses.field(default_factory=list)
    variables: list[OrganizationVariable] = dataclasses.field(default_factory=list)
    rulesets: list[OrganizationRuleset] = dataclasses.field(default_factory=list)
    repositories: list[Repository] = dataclasses.field(default_factory=list)

    _secrets_resolved: bool = False

    @property
    def secrets_resolved(self) -> bool:
        return self._secrets_resolved

    def add_role(self, role: OrganizationRole) -> None:
        self.roles.append(role)

    def get_role(self, name: str) -> OrganizationRole | None:
        return next(filter(lambda x: x.name == name, self.roles), None)  # type: ignore

    def set_roles(self, roles: list[OrganizationRole]) -> None:
        self.roles = roles

    def add_team(self, team: Team) -> None:
        self.teams.append(team)

    def get_team(self, name: str) -> Team | None:
        return next(filter(lambda x: x.name == name, self.teams), None)  # type: ignore

    def set_teams(self, teams: list[Team]) -> None:
        self.teams = teams

    def add_webhook(self, webhook: OrganizationWebhook) -> None:
        self.webhooks.append(webhook)

    def get_webhook(self, url: str) -> OrganizationWebhook | None:
        return next(filter(lambda x: x.url == url, self.webhooks), None)  # type: ignore

    def set_webhooks(self, webhooks: list[OrganizationWebhook]) -> None:
        self.webhooks = webhooks

    def add_secret(self, secret: OrganizationSecret) -> None:
        self.secrets.append(secret)

    def get_secret(self, name: str) -> OrganizationSecret | None:
        return next(filter(lambda x: x.name == name, self.secrets), None)  # type: ignore

    def set_secrets(self, secrets: list[OrganizationSecret]) -> None:
        self.secrets = secrets

    def add_variable(self, variable: OrganizationVariable) -> None:
        self.variables.append(variable)

    def get_variable(self, name: str) -> OrganizationVariable | None:
        return next(filter(lambda x: x.name == name, self.variables), None)  # type: ignore

    def set_variables(self, variables: list[OrganizationVariable]) -> None:
        self.variables = variables

    def add_ruleset(self, ruleset: OrganizationRuleset) -> None:
        self.rulesets.append(ruleset)

    def get_ruleset(self, name: str) -> OrganizationRuleset | None:
        return next(filter(lambda x: x.name == name, self.rulesets), None)  # type: ignore

    def set_rulesets(self, rulesets: list[OrganizationRuleset]) -> None:
        self.rulesets = rulesets

    def add_repository(self, repo: Repository) -> None:
        self.repositories.append(repo)

    def get_repository(self, repo_name: str) -> Repository | None:
        return next(filter(lambda x: x.name == repo_name, self.repositories), None)  # type: ignore

    def set_repositories(self, repos: list[Repository]) -> None:
        self.repositories = repos

    async def validate(
        self,
        config: OtterdogConfig,
        jsonnet_config: JsonnetConfig,
        secret_resolver: SecretResolver,
        provider: GitHubProvider,
    ) -> ValidationContext:
        # only retrieve the list of current organization members if there are teams defined
        if len(self.teams) > 0:
            org_members = {x["login"] for x in await provider.rest_api.org.list_members(self.github_id)}
        else:
            org_members = set()

        default_org = GitHubOrganization.from_model_data(
            jsonnet_config.default_org_config_for_org_id(self.project_name, self.github_id)
        )

        context = ValidationContext(
            self,
            secret_resolver,
            jsonnet_config.template_dir,
            org_members,
            {t.name for t in default_org.teams},
            config.exclude_teams_pattern,
        )
        self.settings.validate(context, self)

        enterprise_plan = self.settings.plan == "enterprise"

        if len(self.roles) > 0 and not enterprise_plan:
            context.add_failure(
                FailureType.ERROR,
                f"use of organization roles requires an 'enterprise' plan, while this organization is "
                f"currently on a '{self.settings.plan}' plan.",
            )
        else:
            for role in self.roles:
                role.validate(context, self)

        for team in self.teams:
            team.validate(context, self)

        for webhook in self.webhooks:
            webhook.validate(context, self)

        for secret in self.secrets:
            secret.validate(context, self)

        if len(self.rulesets) > 0 and not enterprise_plan:
            context.add_failure(
                FailureType.ERROR,
                f"use of organization rulesets requires an 'enterprise' plan, while this organization is "
                f"currently on a '{self.settings.plan}' plan.",
            )
        else:
            for ruleset in self.rulesets:
                ruleset.validate(context, self)

        for repo in self.repositories:
            repo.validate(context, self)

        return context

    @staticmethod
    def _validate_org_config(data: dict[str, Any]) -> None:
        from jsonschema import Draft202012Validator
        from referencing import Registry, Resource
        from referencing.exceptions import NoSuchResource

        with as_file(files(resources).joinpath("schemas")) as resource_dir:

            def retrieve_from_filesystem(uri: str):
                path = resource_dir.joinpath(uri)
                if not path.exists():
                    raise NoSuchResource(ref=uri)  # type: ignore

                contents = json.loads(path.read_text())
                return Resource.from_contents(contents)

            registry = Registry(retrieve=retrieve_from_filesystem)  # type: ignore
            validator = Draft202012Validator(_ORG_SCHEMA, registry=registry)
            validator.validate(data)

    def get_model_objects(self) -> Iterator[tuple[ModelObject, ModelObject | None]]:
        yield self.settings, None
        yield from self.settings.get_model_objects()

        for role in self.roles:
            yield role, None
            yield from role.get_model_objects()

        for team in self.teams:
            yield team, None
            yield from team.get_model_objects()

        for webhook in self.webhooks:
            yield webhook, None
            yield from webhook.get_model_objects()

        for secret in self.secrets:
            yield secret, None
            yield from secret.get_model_objects()

        for variable in self.variables:
            yield variable, None
            yield from variable.get_model_objects()

        for ruleset in self.rulesets:
            yield ruleset, None
            yield from ruleset.get_model_objects()

        for repo in self.repositories:
            yield repo, None
            yield from repo.get_model_objects()

    @classmethod
    def from_model_data(cls, data: dict[str, Any]) -> GitHubOrganization:
        # validate the input data with the json schema.
        cls._validate_org_config(data)

        mapping = {
            "project_name": S("project_name"),
            "github_id": S("github_id"),
            "settings": S("settings") >> F(lambda x: OrganizationSettings.from_model_data(x)),
            "roles": OptionalS("roles", default=[]) >> Forall(lambda x: OrganizationRole.from_model_data(x)),
            "teams": OptionalS("teams", default=[]) >> Forall(lambda x: Team.from_model_data(x)),
            "webhooks": OptionalS("webhooks", default=[]) >> Forall(lambda x: OrganizationWebhook.from_model_data(x)),
            "secrets": OptionalS("secrets", default=[]) >> Forall(lambda x: OrganizationSecret.from_model_data(x)),
            "variables": OptionalS("variables", default=[])
            >> Forall(lambda x: OrganizationVariable.from_model_data(x)),
            "rulesets": OptionalS("rulesets", default=[]) >> Forall(lambda x: OrganizationRuleset.from_model_data(x)),
            "repositories": OptionalS("repositories", default=[]) >> Forall(lambda x: Repository.from_model_data(x)),
        }

        org = cls(**bend(mapping, data))
        org.repositories = [x.coerce_from_org_settings(org.settings) for x in org.repositories]
        return org

    def resolve_secrets(self, secret_resolver: Callable[[str], str]) -> None:
        for webhook in self.webhooks:
            webhook.resolve_secrets(secret_resolver)

        for secret in self.secrets:
            secret.resolve_secrets(secret_resolver)

        for repo in self.repositories:
            repo.resolve_secrets(secret_resolver)

        self._secrets_resolved = True

    def copy_secrets(self, other_org: GitHubOrganization) -> None:
        for webhook in self.webhooks:
            other_webhook = other_org.get_webhook(webhook.url)
            if other_webhook is not None:
                webhook.copy_secrets(other_webhook)

        for secret in self.secrets:
            other_secret = other_org.get_secret(secret.name)
            if other_secret is not None:
                secret.copy_secrets(other_secret)

        for repo in self.repositories:
            other_repo = other_org.get_repository(repo.name)
            if other_repo is not None:
                repo.copy_secrets(other_repo)

    def update_dummy_secrets(self, new_value: str) -> None:
        for model_object, _ in self.get_model_objects():
            model_object.update_dummy_secrets(new_value)

    def unset_settings_requiring_web_ui(self) -> None:
        for model_object, _ in self.get_model_objects():
            model_object.unset_settings_requiring_web_ui()

    def to_jsonnet(self, config: JsonnetConfig, context: PatchContext) -> str:
        default_org = GitHubOrganization.from_model_data(
            config.default_org_config_for_org_id(self.project_name, self.github_id)
        )

        output = StringIO()
        printer = IndentingPrinter(output)

        printer.println(f"local orgs = {config.import_statement};")
        printer.println()
        printer.println(f"orgs.{config.create_org}('{self.project_name}', '{self.github_id}') {{")
        printer.level_up()

        # print organization settings
        printer.print("settings+:")
        self.settings.to_jsonnet(printer, config, context, False, default_org.settings)

        # print organization roles
        if len(self.roles) > 0:
            default_org_role = OrganizationRole.from_model_data(config.default_org_role_config)

            printer.println("roles+: [")
            printer.level_up()

            for role in self.roles:
                role.to_jsonnet(printer, config, context, False, default_org_role)

            printer.level_down()
            printer.println("],")

        # print teams
        if len(self.teams) > 0:
            teams_by_name = associate_by_key(self.teams, lambda x: x.name)
            default_teams_by_name = associate_by_key(default_org.teams, lambda x: x.name)

            # remove teams inherited from the default config
            for default_team_name in set(default_teams_by_name):
                if default_team_name in teams_by_name:
                    teams_by_name.pop(default_team_name)

            if len(teams_by_name) > 0:
                default_team = Team.from_model_data(config.default_team_config)

                printer.println("teams+: [")
                printer.level_up()

                for _, team in sorted(teams_by_name.items()):
                    team.to_jsonnet(printer, config, context, False, default_team)

                printer.level_down()
                printer.println("],")

        # print organization webhooks
        if len(self.webhooks) > 0:
            default_org_webhook = OrganizationWebhook.from_model_data(config.default_org_webhook_config)

            printer.println("webhooks+: [")
            printer.level_up()

            for webhook in self.webhooks:
                webhook.to_jsonnet(printer, config, context, False, default_org_webhook)

            printer.level_down()
            printer.println("],")

        # print organization secrets
        if len(self.secrets) > 0:
            default_org_secret = OrganizationSecret.from_model_data(config.default_org_secret_config)

            printer.println("secrets+: [")
            printer.level_up()

            for secret in self.secrets:
                secret.to_jsonnet(printer, config, context, False, default_org_secret)

            printer.level_down()
            printer.println("],")

        # print organization variables
        if len(self.variables) > 0:
            default_org_variable = OrganizationVariable.from_model_data(config.default_org_variable_config)

            printer.println("variables+: [")
            printer.level_up()

            for variable in self.variables:
                variable.to_jsonnet(printer, config, context, False, default_org_variable)

            printer.level_down()
            printer.println("],")

        # print organization rulesets
        if len(self.rulesets) > 0:
            default_org_ruleset = OrganizationRuleset.from_model_data(config.default_org_ruleset_config)

            printer.println("rulesets+: [")
            printer.level_up()

            for ruleset in self.rulesets:
                ruleset.to_jsonnet(printer, config, context, False, default_org_ruleset)

            printer.level_down()
            printer.println("],")

        # print repositories
        if len(self.repositories) > 0:
            repos_by_name = associate_by_key(self.repositories, lambda x: x.name)
            default_repos_by_name = associate_by_key(default_org.repositories, lambda x: x.name)

            # add all default repos which are not yet contained in repos
            for default_repo_name, default_repo in default_repos_by_name.items():
                if repos_by_name.get(default_repo_name) is None:
                    repos_by_name[default_repo_name] = default_repo

            default_org_repo = Repository.from_model_data(config.default_repo_config)

            printer.println("_repositories+:: [")
            printer.level_up()

            for repo_name, repo in sorted(repos_by_name.items()):
                if repo_name in default_repos_by_name:
                    other_repo = default_repos_by_name[repo_name]
                    extend = True
                else:
                    other_repo = default_org_repo
                    extend = False

                repo.to_jsonnet(printer, config, context, extend, other_repo)

            printer.level_down()
            printer.println("],")

        printer.level_down()
        printer.println("}")

        return output.getvalue()

    def generate_live_patch(
        self, current_organization: GitHubOrganization, context: LivePatchContext, handler: LivePatchHandler
    ) -> None:
        OrganizationRole.generate_live_patch_of_list(self.roles, current_organization.roles, None, context, handler)
        Team.generate_live_patch_of_list(self.teams, current_organization.teams, None, context, handler)
        OrganizationSettings.generate_live_patch(self.settings, current_organization.settings, None, context, handler)
        OrganizationWebhook.generate_live_patch_of_list(
            self.webhooks, current_organization.webhooks, None, context, handler
        )
        OrganizationSecret.generate_live_patch_of_list(
            self.secrets, current_organization.secrets, None, context, handler
        )
        OrganizationVariable.generate_live_patch_of_list(
            self.variables, current_organization.variables, None, context, handler
        )
        OrganizationRuleset.generate_live_patch_of_list(
            self.rulesets, current_organization.rulesets, None, context, handler
        )
        Repository.generate_live_patch_of_list(
            self.repositories, current_organization.repositories, None, context, handler
        )

    @classmethod
    def load_from_file(cls, github_id: str, config_file: str) -> GitHubOrganization:
        if not os.path.exists(config_file):
            msg = f"configuration file '{config_file}' for organization '{github_id}' does not exist"
            raise RuntimeError(msg)

        _logger.debug("loading configuration for organization '%s' from file '%s'", github_id, config_file)
        data = jsonnet_evaluate_file(config_file)

        return cls.from_model_data(data)

    @classmethod
    async def load_from_provider(
        cls,
        project_name: str,
        github_id: str,
        jsonnet_config: JsonnetConfig,
        provider: GitHubProvider,
        no_web_ui: bool = False,
        concurrency: int | None = None,
        repo_filter: str | None = None,
        exclude_teams: Pattern | None = None,
    ) -> GitHubOrganization:
        import asyncer

        @debug_times("settings")
        async def _load_settings() -> OrganizationSettings:
            # FIXME: this uses the keys from the model schema which might be different to the provider schema
            #        for now this is the same for organization settings, but there might be cases where it is different.
            default_settings = jsonnet_config.default_org_config["settings"]
            included_keys = set(default_settings.keys())
            github_settings = await provider.get_org_settings(github_id, included_keys, no_web_ui)

            if "workflows" in included_keys:
                github_settings["workflows"] = await provider.get_org_workflow_settings(github_id)

            settings = OrganizationSettings.from_provider_data(github_id, github_settings)

            if "workflows" in included_keys:
                github_org_workflow_data = await provider.get_org_workflow_settings(github_id)
                settings.workflows = OrganizationWorkflowSettings.from_provider_data(
                    github_id, github_org_workflow_data
                )

            if "custom_properties" in included_keys:
                github_custom_properties = await provider.get_org_custom_properties(github_id)
                settings.custom_properties = [
                    CustomProperty.from_provider_data(github_id, x) for x in github_custom_properties
                ]

            return settings

        org_settings = await _load_settings()
        org = cls(project_name, github_id, org_settings)

        @debug_times("roles")
        async def _load_roles() -> None:
            if jsonnet_config.default_org_role_config is not None and org.settings.plan == "enterprise":
                github_roles = await provider.get_org_custom_roles(github_id)
                for role in github_roles:
                    org.add_role(OrganizationRole.from_provider_data(github_id, role))
            else:
                _logger.debug("not reading org webhooks, no default config available")

        @debug_times("teams")
        async def _load_teams() -> None:
            if jsonnet_config.default_team_config is not None:
                default_org = GitHubOrganization.from_model_data(
                    jsonnet_config.default_org_config_for_org_id(project_name, github_id)
                )

                github_teams = await provider.get_org_teams(github_id)
                for team in github_teams:
                    team_name = team["name"]
                    team_slug = team["slug"]

                    default_org.get_team(team_name)
                    if (
                        exclude_teams is not None
                        and exclude_teams.match(team_slug)
                        and default_org.get_team(team_name) is None
                    ):
                        continue
                    team_members = await provider.get_org_team_members(github_id, team_slug)
                    team["members"] = team_members
                    org.add_team(Team.from_provider_data(github_id, team))
            else:
                _logger.debug("not reading teams, no default config available")

        @debug_times("webhooks")
        async def _load_webhooks() -> None:
            if jsonnet_config.default_org_webhook_config is not None:
                github_webhooks = await provider.get_org_webhooks(github_id)
                for webhook in github_webhooks:
                    org.add_webhook(OrganizationWebhook.from_provider_data(github_id, webhook))
            else:
                _logger.debug("not reading org webhooks, no default config available")

        @debug_times("secrets")
        async def _load_secrets() -> None:
            if jsonnet_config.default_org_secret_config is not None:
                github_secrets = await provider.get_org_secrets(github_id)
                for secret in github_secrets:
                    org.add_secret(OrganizationSecret.from_provider_data(github_id, secret))
            else:
                _logger.debug("not reading org secrets, no default config available")

        @debug_times("variables")
        async def _load_variables() -> None:
            if jsonnet_config.default_org_variable_config is not None:
                github_variables = await provider.get_org_variables(github_id)
                for variable in github_variables:
                    org.add_variable(OrganizationVariable.from_provider_data(github_id, variable))
            else:
                _logger.debug("not reading org secrets, no default config available")

        @debug_times("rulesets")
        async def _load_rulesets() -> None:
            if jsonnet_config.default_org_ruleset_config is not None and org.settings.plan == "enterprise":
                github_rulesets = await provider.get_org_rulesets(github_id)
                for ruleset in github_rulesets:
                    org.add_ruleset(OrganizationRuleset.from_provider_data(github_id, ruleset))
            else:
                _logger.debug("not reading org secrets, no default config available")

        @debug_times("repos")
        async def _load_repos() -> None:
            if jsonnet_config.default_repo_config is not None:
                async for repo in _load_repos_from_provider(
                    github_id,
                    org.settings,
                    provider,
                    jsonnet_config,
                    concurrency,
                    repo_filter,
                ):
                    org.add_repository(repo)
            else:
                _logger.debug("not reading repos, no default config available")

        async with asyncer.create_task_group() as task_group:
            task_group.soonify(_load_roles)()
            task_group.soonify(_load_teams)()
            task_group.soonify(_load_webhooks)()
            task_group.soonify(_load_secrets)()
            task_group.soonify(_load_variables)()
            task_group.soonify(_load_rulesets)()
            task_group.soonify(_load_repos)()

        return org


async def _process_single_repo(
    gh_client: GitHubProvider,
    github_id: str,
    org_settings: OrganizationSettings,
    repo_name: str,
    jsonnet_config: JsonnetConfig,
    teams: dict[str, Any],
    app_installations: dict[str, str],
) -> tuple[str, Repository]:
    rest_api = gh_client.rest_api

    # get repo data
    github_repo_data = await rest_api.repo.get_repo_data(github_id, repo_name)
    repo = Repository.from_provider_data(github_id, github_repo_data)

    github_repo_workflow_data = await rest_api.repo.get_workflow_settings(github_id, repo_name)
    repo.workflows = RepositoryWorkflowSettings.from_provider_data(github_id, github_repo_workflow_data)

    if jsonnet_config.default_branch_protection_rule_config is not None:
        # get branch protection rules of the repo
        rules = await gh_client.get_branch_protection_rules(github_id, repo_name)
        for github_rule in rules:
            repo.add_branch_protection_rule(BranchProtectionRule.from_provider_data(github_id, github_rule))
    else:
        _logger.debug("not reading branch protection rules, no default config available")

    if jsonnet_config.default_repo_ruleset_config is not None and (
        repo.private is False or org_settings.plan == "enterprise"
    ):
        # get rulesets of the repo
        rulesets = await rest_api.repo.get_rulesets(github_id, repo_name)
        for github_ruleset in rulesets:
            # FIXME: need to associate an app id to its slug
            #        GitHub does not support that atm, so we lookup the currently installed
            #        apps for an organization which provide a mapping from id to slug.
            for actor in github_ruleset.get("bypass_actors", []):
                if actor.get("actor_type", None) == "Integration":
                    actor_id = str(actor.get("actor_id", 0))
                    if actor_id in app_installations:
                        actor["app_slug"] = app_installations[actor_id]
                elif actor.get("actor_type", None) == "Team":
                    actor_id = str(actor.get("actor_id", 0))
                    if actor_id in teams:
                        actor["team_slug"] = teams[actor_id]

            for rule in github_ruleset.get("rules", []):
                if rule.get("type", None) == "required_status_checks":
                    required_status_checks = rule.get("parameters", {}).get("required_status_checks", [])
                    for status_check in required_status_checks:
                        integration_id = str(status_check.get("integration_id", 0))
                        if integration_id in app_installations:
                            status_check["app_slug"] = app_installations[integration_id]

            repo.add_ruleset(RepositoryRuleset.from_provider_data(github_id, github_ruleset))
    else:
        _logger.debug("not reading repo rulesets, no default config available")

    if jsonnet_config.default_org_webhook_config is not None:
        # get webhooks of the repo
        webhooks = await rest_api.repo.get_webhooks(github_id, repo_name)
        for github_webhook in webhooks:
            repo.add_webhook(RepositoryWebhook.from_provider_data(github_id, github_webhook))
    else:
        _logger.debug("not reading repo webhooks, no default config available")

    if jsonnet_config.default_repo_secret_config is not None:
        # get secrets of the repo
        secrets = await rest_api.repo.get_secrets(github_id, repo_name)
        for github_secret in secrets:
            repo.add_secret(RepositorySecret.from_provider_data(github_id, github_secret))
    else:
        _logger.debug("not reading repo secrets, no default config available")

    if jsonnet_config.default_repo_variable_config is not None:
        # get variables of the repo
        variables = await rest_api.repo.get_variables(github_id, repo_name)
        for github_variable in variables:
            repo.add_variable(RepositoryVariable.from_provider_data(github_id, github_variable))
    else:
        _logger.debug("not reading repo variables, no default config available")

    if jsonnet_config.default_environment_config is not None:
        # get environments of the repo
        environments = await rest_api.repo.get_environments(github_id, repo_name)
        for github_environment in environments:
            repo.add_environment(Environment.from_provider_data(github_id, github_environment))
    else:
        _logger.debug("not reading environments, no default config available")

    _logger.debug("done retrieving data for repo '%s'", repo_name)

    return repo_name, repo


async def _load_repos_from_provider(
    github_id: str,
    org_settings: OrganizationSettings,
    provider: GitHubProvider,
    jsonnet_config: JsonnetConfig,
    concurrency: int | None = None,
    repo_filter: str | None = None,
) -> AsyncIterator[Repository]:
    import fnmatch

    repo_names = await provider.get_repos(github_id)

    if repo_filter is not None:
        repo_names = fnmatch.filter(repo_names, repo_filter)

    teams = {str(team["id"]): f"{github_id}/{team['slug']}" for team in await provider.get_org_teams(github_id)}

    app_installations = {
        str(installation["app_id"]): installation["app_slug"]
        for installation in await provider.rest_api.org.get_app_installations(github_id)
    }

    # limit the number of repos that are processed concurrently to avoid hitting secondary rate limits
    sem = asyncio.Semaphore(50 if concurrency is None else concurrency)

    async def safe_process(repo_name):
        async with sem:
            return await _process_single_repo(
                provider,
                github_id,
                org_settings,
                repo_name,
                jsonnet_config,
                teams,
                app_installations,
            )

    if concurrency is not None:
        chunk_size = 50
        result = []
        for chunk in divide_chunks(repo_names, chunk_size):
            result.extend(await asyncio.gather(*[safe_process(repo_name) for repo_name in chunk]))

            # after processing a full chunk, wait for 30s to avoid hitting secondary rate limits
            if len(chunk) == chunk_size:
                await asyncio.sleep(30)
    else:
        result = await asyncio.gather(*[safe_process(repo_name) for repo_name in repo_names])

    for data in result:
        _, repo_data = data
        yield repo_data


def divide_chunks(input_list, n):
    # looping till length of input_list
    for i in range(0, len(input_list), n):
        yield input_list[i : i + n]



#  *******************************************************************************
#  Copyright (c) 2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from datetime import timedelta

import pytest

from otterdog.webapp.utils import backoff_if_needed, current_utc_time


@pytest.mark.asyncio
async def test_backoff_if_needed():
    # check that we wait the required timeout period
    start = current_utc_time()
    await backoff_if_needed(start, timedelta(seconds=3))
    end = current_utc_time()
    assert end - start > timedelta(seconds=3)

    # check that we do not wait if the required timeout already expired
    start = current_utc_time()
    await backoff_if_needed(start - timedelta(seconds=60), timedelta(seconds=3))
    end = current_utc_time()
    assert end - start < timedelta(seconds=1)



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

from typing import TYPE_CHECKING

from aiofiles import ospath

from otterdog.models.github_organization import GitHubOrganization
from otterdog.utils import unwrap

from .apply import ApplyOperation

if TYPE_CHECKING:
    from otterdog.jsonnet import JsonnetConfig


class LocalApplyOperation(ApplyOperation):
    def __init__(
        self,
        suffix: str,
        force_processing: bool,
        no_web_ui: bool,
        repo_filter: str,
        update_webhooks: bool,
        update_secrets: bool,
        update_filter: str,
        delete_resources: bool,
        resolve_secrets: bool = True,
        include_resources_with_secrets: bool = True,
    ) -> None:
        super().__init__(
            force_processing=force_processing,
            no_web_ui=no_web_ui,
            repo_filter=repo_filter,
            update_webhooks=update_webhooks,
            update_secrets=update_secrets,
            update_filter=update_filter,
            delete_resources=delete_resources,
            resolve_secrets=resolve_secrets,
            include_resources_with_secrets=include_resources_with_secrets,
        )

        self._suffix = suffix
        self._other_org: GitHubOrganization | None = None

    @property
    def suffix(self) -> str:
        return self._suffix

    @property
    def other_org(self) -> GitHubOrganization:
        return unwrap(self._other_org)

    def pre_execute(self) -> None:
        self.printer.println("Applying local changes:")
        self.print_legend()

    def verbose_output(self):
        return False

    def coerce_current_org(self) -> bool:
        return True

    async def load_current_org(
        self, project_name: str, github_id: str, jsonnet_config: JsonnetConfig
    ) -> GitHubOrganization:
        other_org_file_name = jsonnet_config.org_config_file + self.suffix

        if not await ospath.exists(other_org_file_name):
            raise RuntimeError(f"configuration file '{other_org_file_name}' does not exist")

        github_organization = GitHubOrganization.load_from_file(github_id, other_org_file_name)

        if self.no_web_ui is True:
            github_organization.unset_settings_requiring_web_ui()

        return github_organization



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

from typing import TYPE_CHECKING

from rich import box
from rich.table import Table

from . import Operation

if TYPE_CHECKING:
    from otterdog.config import OrganizationConfig


class ListProjectsOperation(Operation):
    """
    Lists projects and their associated GitHub ID.
    """

    def __init__(self):
        super().__init__()
        self._table = Table(title="Projects", box=box.ROUNDED)

    @property
    def table(self) -> Table:
        return self._table

    def pre_execute(self) -> None:
        self.table.add_column("Project name", justify="left", style="cyan", no_wrap=True)
        self.table.add_column("GitHub ID", style="magenta")
        self.table.add_column("Index", justify="right", style="green")

    def post_execute(self) -> None:
        self.printer.console.print(self.table)

    async def execute(
        self,
        org_config: OrganizationConfig,
        org_index: int | None = None,
        org_count: int | None = None,
    ) -> int:
        self.table.add_row(org_config.name, org_config.github_id, str(org_index or -1))
        return 0



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

import abc
import dataclasses
from typing import Any, TypeVar

from jsonbender import F, OptionalS  # type: ignore

from otterdog.models import (
    FailureType,
    ModelObject,
    ValidationContext,
)
from otterdog.utils import is_set_and_valid

RT = TypeVar("RT", bound="Role")


@dataclasses.dataclass
class Role(ModelObject, abc.ABC):
    """
    Represents a Role.
    """

    id: int = dataclasses.field(metadata={"external_only": True})
    source: str = dataclasses.field(metadata={"external_only": True})
    name: str = dataclasses.field(metadata={"key": True})
    description: str
    permissions: list[str]
    base_role: str

    def validate(self, context: ValidationContext, parent_object: Any) -> None:
        if is_set_and_valid(self.base_role):
            if self.base_role not in {"none", "read", "triage", "write", "maintain", "admin"}:
                context.add_failure(
                    FailureType.ERROR,
                    f"{self.get_model_header(parent_object)} has 'base_role' of value '{self.base_role}', "
                    f"while only values ('none' | 'read' | 'triage' | 'write' | 'maintain' | 'admin') are allowed.",
                )

        if is_set_and_valid(self.base_role) and is_set_and_valid(self.permissions):
            if self.base_role == "none" and len(self.permissions) > 0:
                context.add_failure(
                    FailureType.ERROR,
                    f"{self.get_model_header(parent_object)} has 'base_role' of value '{self.base_role}', "
                    f"and specified additional permissions, which is not allowed.",
                )

        # TODO: add validation for allowed permissions

    @classmethod
    def get_mapping_from_provider(cls, org_id: str, data: dict[str, Any]) -> dict[str, Any]:
        mapping = super().get_mapping_from_provider(org_id, data)

        mapping.update(
            {
                "base_role": OptionalS("base_role") >> F(lambda x: x if x is not None else "none"),
            }
        )
        return mapping



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING

from aiofiles import open, tempfile

from otterdog.models import PatchContext
from otterdog.models.github_organization import GitHubOrganization
from otterdog.utils import sort_jsonnet, strip_trailing_commas

from . import Operation

if TYPE_CHECKING:
    from otterdog.config import OrganizationConfig


class CanonicalDiffOperation(Operation):
    """
    Performs a canonical diff for organization configurations, i.e. comparing the current configuration
    with a canonical version and showing the unified diff of it.
    """

    def __init__(self):
        super().__init__()

    def pre_execute(self) -> None:
        self.printer.println("Showing canonical diff:")

    async def execute(
        self,
        org_config: OrganizationConfig,
        org_index: int | None = None,
        org_count: int | None = None,
    ) -> int:
        github_id = org_config.github_id
        jsonnet_config = org_config.jsonnet_config
        await jsonnet_config.init_template()

        self._print_project_header(org_config, org_index, org_count)

        org_file_name = jsonnet_config.org_config_file
        if not await self.check_config_file_exists(org_file_name):
            return 1

        try:
            organization = GitHubOrganization.load_from_file(github_id, org_file_name)
        except RuntimeError as ex:
            self.printer.print_error(f"failed to load configuration: {ex!s}")
            return 1

        async with open(org_file_name) as file:
            original_config = await file.read()

        original_config_without_comments: list[str] = strip_trailing_commas(
            sort_jsonnet(list(filter(lambda x: not x.strip().startswith("#"), original_config.split("\n"))))
        )

        context = PatchContext(github_id, organization.settings)
        canonical_config = organization.to_jsonnet(jsonnet_config, context)
        canonical_config_as_lines = strip_trailing_commas(sort_jsonnet(canonical_config.split("\n")))

        self.printer.println()

        async for line in self._diff(
            canonical_config_as_lines, original_config_without_comments, "canonical", "original"
        ):
            if line.startswith("+"):
                self.printer.println(f"[green]{line}[/]")
            elif line.startswith("-"):
                self.printer.println(f"[red]{line}[/]")
            else:
                self.printer.println(line)

        return 0

    @staticmethod
    async def _diff(a, b, name_a, name_b):
        async with tempfile.NamedTemporaryFile() as file:
            await file.write(bytes("\n".join(a), "utf-8"))
            await file.flush()

            try:
                cmd = f"diff --label {name_a} --label {name_b} -u -w - {file.name}"
                proc = await asyncio.create_subprocess_shell(
                    cmd,
                    stdin=asyncio.subprocess.PIPE,
                    stdout=asyncio.subprocess.PIPE,
                    stderr=asyncio.subprocess.PIPE,
                )

                out, _ = await proc.communicate(bytes("\n".join(b), "utf-8"))

                for line in out.decode("utf-8").split("\n"):
                    yield line
            except BrokenPipeError as ex:
                raise RuntimeError(f"failed to run diff command: {ex!s}") from ex



#  *******************************************************************************
#  Copyright (c) 2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

import dataclasses


@dataclasses.dataclass
class RequestStatistics:
    total_requests: int = 0
    cached_responses: int = 0
    remaining_rate_limit: int = -1

    def merge(self, other: RequestStatistics) -> None:
        self.total_requests += other.total_requests
        self.cached_responses += other.cached_responses

        if self.remaining_rate_limit == -1:
            self.remaining_rate_limit = other.remaining_rate_limit
        elif other.remaining_rate_limit == -1:
            self.remaining_rate_limit = self.remaining_rate_limit
        else:
            self.remaining_rate_limit = min(self.remaining_rate_limit, other.remaining_rate_limit)

    def sent_request(self) -> None:
        self.total_requests += 1

    def received_cached_response(self) -> None:
        self.cached_responses += 1

    def update_remaining_rate_limit(self, remaining: int) -> None:
        self.remaining_rate_limit = remaining



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import base64
from typing import Any

from otterdog.logging import get_logger
from otterdog.providers.github.exception import GitHubException

from . import RestApi, RestClient

_logger = get_logger(__name__)


class ContentClient(RestClient):
    def __init__(self, rest_api: RestApi):
        super().__init__(rest_api)

    async def get_contents(
        self, org_id: str, repo_name: str, path: str, ref: str | None = None
    ) -> list[dict[str, Any]]:
        json_response = await self.get_content_object(org_id, repo_name, path, ref)
        if not isinstance(json_response, list):
            raise RuntimeError(f"unexpected result for retrieving contents of path '{path}': '{type(json_response)}'")
        return json_response

    async def get_content_object(
        self, org_id: str, repo_name: str, path: str, ref: str | None = None
    ) -> dict[str, Any] | list[dict]:
        _logger.debug("retrieving content '%s' from repo '%s/%s'", path, org_id, repo_name)

        try:
            params = {"ref": ref} if ref is not None else None

            return await self.requester.request_json(
                "GET", f"/repos/{org_id}/{repo_name}/contents/{path}", params=params
            )
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving content '{path}' from repo '{repo_name}':\n{ex}") from ex

    async def get_content(self, org_id: str, repo_name: str, path: str, ref: str | None = None) -> str:
        json_response = await self.get_content_object(org_id, repo_name, path, ref)
        if not isinstance(json_response, dict):
            raise RuntimeError(f"unexpected result for retrieving content of path '{path}': '{type(json_response)}'")
        return base64.b64decode(json_response["content"]).decode("utf-8")

    async def get_content_with_sha(
        self, org_id: str, repo_name: str, path: str, ref: str | None = None
    ) -> tuple[str, str]:
        json_response = await self.get_content_object(org_id, repo_name, path, ref)
        if not isinstance(json_response, dict):
            raise RuntimeError(f"unexpected result for retrieving content of path '{path}': '{type(json_response)}'")
        return base64.b64decode(json_response["content"]).decode("utf-8"), json_response["sha"]

    async def update_content(
        self,
        org_id: str,
        repo_name: str,
        path: str,
        content: str,
        ref: str | None = None,
        message: str | None = None,
        author_name: str | None = None,
        author_email: str | None = None,
        author_is_committer: bool = False,
    ) -> bool:
        _logger.debug("putting content '%s' to repo '%s/%s'", path, org_id, repo_name)

        try:
            json_response = await self.get_content_object(org_id, repo_name, path, ref)
            if not isinstance(json_response, dict):
                raise RuntimeError(
                    f"unexpected result for retrieving content of path '{path}': '{type(json_response)}'"
                )
            old_sha = json_response["sha"]
            old_content = base64.b64decode(json_response["content"]).decode("utf-8")
        except RuntimeError:
            old_sha = None
            old_content = None

        # check if the content has changed, otherwise do not update
        if old_content is not None and content == old_content:
            _logger.debug("not updating content, no changes")
            return False

        base64_encoded_data = base64.b64encode(content.encode("utf-8"))
        base64_content = base64_encoded_data.decode("utf-8")

        push_message = f"Updating file '{path}' with otterdog." if message is None else message

        data: dict[str, Any] = {
            "message": push_message,
            "content": base64_content,
        }

        if ref is not None:
            data["branch"] = ref

        if author_name is not None:
            author = {
                "name": author_name,
                "email": author_email if author_email is not None else "",
            }
            data["author"] = author

            if author_is_committer is True:
                data["committer"] = author

        if old_sha is not None:
            data["sha"] = old_sha

        try:
            await self.requester.request_json("PUT", f"/repos/{org_id}/{repo_name}/contents/{path}", data)
            return True
        except GitHubException as ex:
            raise RuntimeError(f"failed putting content '{path}' to repo '{repo_name}':\n{ex}") from ex

    async def delete_content(
        self,
        org_id: str,
        repo_name: str,
        path: str,
        message: str | None = None,
    ) -> bool:
        _logger.debug("deleting content '%s' in repo '%s/%s'", path, org_id, repo_name)

        try:
            json_response = await self.get_content_object(org_id, repo_name, path)
            if not isinstance(json_response, dict):
                raise RuntimeError(
                    f"unexpected result for retrieving content of path '{path}': '{type(json_response)}'"
                )
            old_sha = json_response["sha"]
        except RuntimeError:
            old_sha = None

        if old_sha is None:
            return False

        push_message = f"Deleting file '{path}' with otterdog." if message is None else message

        data = {"message": push_message, "sha": old_sha}

        try:
            await self.requester.request_json("DELETE", f"/repos/{org_id}/{repo_name}/contents/{path}", data)
            return True
        except GitHubException as ex:
            raise RuntimeError(f"failed deleting content '{path}' in repo '{repo_name}':\n{ex}") from ex



#  *******************************************************************************
#  Copyright (c) 2023-2025 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from collections.abc import Mapping
from typing import Any

from otterdog.jsonnet import JsonnetConfig
from otterdog.models import ModelObject
from otterdog.models.repository import Repository
from otterdog.utils import UNSET, Change, query_json

from . import ModelTest


class RepositoryTest(ModelTest):
    def create_model(self, data: Mapping[str, Any]) -> ModelObject:
        return Repository.from_model_data(data)

    @property
    def template_function(self) -> str:
        return JsonnetConfig.create_repo

    @property
    def model_data(self):
        return self.load_json_resource("otterdog-repo.json")

    @property
    def provider_data(self):
        return self.load_json_resource("github-repo.json")

    def test_load_from_model(self):
        repo = Repository.from_model_data(self.model_data)

        assert repo.name == "otterdog-defaults"
        assert repo.description is None
        assert repo.homepage is None
        assert repo.private is False
        assert repo.has_issues is True
        assert repo.has_projects is True
        assert repo.has_wiki is True
        assert repo.default_branch == "main"
        assert repo.allow_rebase_merge is True
        assert repo.allow_merge_commit is True
        assert repo.allow_squash_merge is True
        assert repo.allow_auto_merge is False
        assert repo.delete_branch_on_merge is False
        assert repo.allow_update_branch is False
        assert repo.squash_merge_commit_title == "COMMIT_OR_PR_TITLE"
        assert repo.squash_merge_commit_message == "COMMIT_MESSAGES"
        assert repo.merge_commit_title == "MERGE_MESSAGE"
        assert repo.merge_commit_message == "PR_TITLE"
        assert repo.archived is False
        assert repo.allow_forking is True
        assert repo.web_commit_signoff_required is False
        assert repo.secret_scanning == "enabled"
        assert repo.secret_scanning_push_protection is UNSET
        assert repo.dependabot_alerts_enabled is True

        assert repo.aliases == ["oldname"]
        assert repo.post_process_template_content == []
        assert repo.auto_init is False

    def test_load_from_provider(self):
        repo = Repository.from_provider_data(self.org_id, self.provider_data)

        assert repo.id == 605555050
        assert repo.node_id == "R_kgDOJBgJag"
        assert repo.name == "otterdog-defaults"
        assert repo.description is None
        assert repo.homepage is None
        assert repo.private is False
        assert repo.has_issues is True
        assert repo.has_projects is True
        assert repo.has_wiki is True
        assert repo.default_branch == "main"
        assert repo.allow_rebase_merge is True
        assert repo.allow_merge_commit is True
        assert repo.allow_squash_merge is True
        assert repo.allow_auto_merge is False
        assert repo.delete_branch_on_merge is False
        assert repo.allow_update_branch is False
        assert repo.squash_merge_commit_title == "COMMIT_OR_PR_TITLE"
        assert repo.squash_merge_commit_message == "COMMIT_MESSAGES"
        assert repo.merge_commit_title == "MERGE_MESSAGE"
        assert repo.merge_commit_message == "PR_TITLE"
        assert repo.archived is False
        assert repo.allow_forking is True
        assert repo.web_commit_signoff_required is False
        assert repo.secret_scanning == "enabled"
        assert repo.secret_scanning_push_protection == "disabled"
        assert repo.dependabot_alerts_enabled is True

    async def test_to_provider(self):
        repo = Repository.from_model_data(self.model_data)

        repo.description = UNSET

        provider_data = await repo.to_provider_data(self.org_id, self.provider)

        assert len(provider_data) == 22
        assert provider_data["name"] == "otterdog-defaults"
        assert provider_data.get("description") is None

        assert query_json("security_and_analysis.secret_scanning.status", provider_data) or "" == "enabled"

    async def test_changes_to_provider(self):
        current = Repository.from_model_data(self.model_data)
        other = Repository.from_model_data(self.model_data)

        other.name = "other"
        other.has_wiki = False
        other.secret_scanning = "disabled"

        changes = current.get_difference_from(other)
        provider_data = await Repository.changes_to_provider(self.org_id, changes, self.provider)

        assert len(provider_data) == 3
        assert provider_data["name"] == "otterdog-defaults"
        assert provider_data["has_wiki"] is True
        assert query_json("security_and_analysis.secret_scanning.status", provider_data) or "" == "enabled"

    def test_patch(self):
        current = Repository.from_model_data(self.model_data)

        default = Repository.from_model_data(self.model_data)

        default.name = None
        default.web_commit_signoff_required = True

        patch = current.get_patch_to(default)

        assert len(patch) == 2
        assert patch["name"] == current.name
        assert patch["web_commit_signoff_required"] is current.web_commit_signoff_required

    def test_difference(self):
        current = Repository.from_model_data(self.model_data)
        other = Repository.from_model_data(self.model_data)

        other.name = "other"
        other.has_wiki = False

        diff = current.get_difference_from(other)

        assert len(diff) == 2
        assert diff["name"] == Change(other.name, current.name)
        assert diff["has_wiki"] == Change(other.has_wiki, current.has_wiki)



#  *******************************************************************************
#  Copyright (c) 2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import unittest
from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from parameterized import parameterized  # type: ignore

from otterdog.webapp.webhook.comment_handlers import (
    ApplyCommentHandler,
    CheckSyncCommentHandler,
    CommentHandler,
    DoneCommentHandler,
    HelpCommentHandler,
    MergeCommentHandler,
    TeamInfoCommentHandler,
    ValidateCommentHandler,
)

T = TypeVar("T", bound=CommentHandler)


class CommentHandlerTest(unittest.TestCase, ABC, Generic[T]):
    @property
    @abstractmethod
    def handler(self) -> CommentHandler:
        pass

    def _test_matches(self, test_input, expected):
        match = self.handler.matches(test_input)
        if expected is True:
            assert match is not None
        else:
            assert match is None


class HelpCommentHandlerTest(CommentHandlerTest[HelpCommentHandler]):
    @property
    def handler(self) -> CommentHandler:
        return HelpCommentHandler()

    @parameterized.expand(
        [
            ("/otterdog help", True),
            ("   /otterdog help   ", False),
            ("/help", False),
            ("/otterdog hel", False),
        ]
    )
    def test_matches(self, test_input, expected):
        self._test_matches(test_input, expected)


class TeamInfoCommentHandlerTest(CommentHandlerTest[TeamInfoCommentHandler]):
    @property
    def handler(self) -> TeamInfoCommentHandler:
        return TeamInfoCommentHandler()

    @parameterized.expand(
        [
            ("/otterdog team-info", True),
            ("   /otterdog team-info   ", False),
            ("/team-info", False),
            ("/otterdog team", False),
        ]
    )
    def test_matches(self, test_input, expected):
        self._test_matches(test_input, expected)


class CheckSyncCommentHandlerTest(CommentHandlerTest[CheckSyncCommentHandler]):
    @property
    def handler(self) -> CommentHandler:
        return CheckSyncCommentHandler()

    @parameterized.expand(
        [
            ("/otterdog check-sync", True),
            ("   /otterdog check-sync   ", False),
            ("/check-sync", False),
            ("/otterdog check", False),
        ]
    )
    def test_matches(self, test_input, expected):
        self._test_matches(test_input, expected)


class ApplyCommentHandlerTest(CommentHandlerTest[ApplyCommentHandler]):
    @property
    def handler(self) -> CommentHandler:
        return ApplyCommentHandler()

    @parameterized.expand(
        [
            ("/otterdog apply", True),
            ("   /otterdog apply   ", False),
            ("/apply", False),
            ("/otterdog appl", False),
        ]
    )
    def test_matches(self, test_input, expected):
        self._test_matches(test_input, expected)


class DoneCommentHandlerTest(CommentHandlerTest[DoneCommentHandler]):
    @property
    def handler(self) -> CommentHandler:
        return DoneCommentHandler()

    @parameterized.expand(
        [
            ("/otterdog done", True),
            ("   /otterdog done   ", False),
            ("/done", False),
            ("/otterdog don", False),
        ]
    )
    def test_matches(self, test_input, expected):
        self._test_matches(test_input, expected)


class MergeCommentHandlerTest(CommentHandlerTest[MergeCommentHandler]):
    @property
    def handler(self) -> CommentHandler:
        return MergeCommentHandler()

    @parameterized.expand(
        [
            ("/otterdog merge", True),
            ("   /otterdog merge   ", False),
            ("/merge", False),
            ("/otterdog mer", False),
        ]
    )
    def test_matches(self, test_input, expected):
        self._test_matches(test_input, expected)


class ValidateCommentHandlerTest(CommentHandlerTest[ValidateCommentHandler]):
    @property
    def handler(self) -> CommentHandler:
        return ValidateCommentHandler()

    @parameterized.expand(
        [
            ("/otterdog validate", True),
            ("/otterdog validate info", True),
            ("   /otterdog validate   ", False),
            ("   /otterdog validate   info", False),
            ("/validate", False),
            ("/otterdog validat", False),
        ]
    )
    def test_matches(self, test_input, expected):
        self._test_matches(test_input, expected)



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import pytest

from otterdog.utils import (
    UNSET,
    camel_to_snake_case,
    deep_merge_dict,
    is_different_ignoring_order,
    is_ghsa_repo,
    parse_template_url,
    patch_to_other,
    snake_to_camel_case,
    snake_to_normal_case,
)


def test_is_different_ignoring_order():
    assert is_different_ignoring_order(1, 2) is True
    assert is_different_ignoring_order(None, None) is False
    assert is_different_ignoring_order(1, None) is True
    assert is_different_ignoring_order(None, 1) is True

    assert is_different_ignoring_order([], []) is False
    assert is_different_ignoring_order([1], [1]) is False
    assert is_different_ignoring_order([1, 2], [2, 1]) is False
    assert is_different_ignoring_order(["Hello"], ["World"]) is True
    assert is_different_ignoring_order(["Hello", "World"], ["World", "Hello"]) is False

    assert is_different_ignoring_order({"a": 1}, {"a": 1}) is False
    assert is_different_ignoring_order({"a": 1}, {"b": 1}) is True
    assert is_different_ignoring_order({"a": 1, "b": 2}, {"b": 2, "a": 1}) is False

    assert is_different_ignoring_order({"a"}, {"a"}) is False
    assert is_different_ignoring_order({"a", "b"}, {"b", "a"}) is False

    assert is_different_ignoring_order(UNSET, UNSET) is False
    assert is_different_ignoring_order(1, UNSET) is True


def test_patch_to_other():
    assert patch_to_other(1, 1) == (False, None)
    assert patch_to_other(1, 2) == (True, 1)

    assert patch_to_other([1], [1]) == (False, None)
    assert patch_to_other([1, 2], [1]) == (True, [2])
    assert patch_to_other([1, 2], [2, 1]) == (False, None)


def test_snake_to_camel_case():
    assert snake_to_camel_case("name") == "name"
    assert snake_to_camel_case("required_status_checks") == "requiredStatusChecks"
    assert snake_to_camel_case("required__status_checks") == "requiredStatusChecks"


def test_snake_to_normal_case():
    assert snake_to_normal_case("name") == "Name"
    assert snake_to_normal_case("required_status_checks") == "Required Status Checks"
    assert snake_to_normal_case("required__status_checks") == "Required Status Checks"


def test_camel_to_snake_case():
    assert camel_to_snake_case("name") == "name"
    assert camel_to_snake_case("requiredStatusChecks") == "required_status_checks"
    assert camel_to_snake_case("RequiredStatusChecks") == "required_status_checks"
    assert camel_to_snake_case("someXYZ") == "some_xyz"


def test_parse_template_url():
    assert parse_template_url("https://github.com/EclipseFdn/otterdog-defaults#otterdog-defaults.libsonnet@main") == (
        "https://github.com/EclipseFdn/otterdog-defaults",
        "otterdog-defaults.libsonnet",
        "main",
    )

    with pytest.raises(ValueError):
        assert parse_template_url("https://github.com/EclipseFdn/otterdog-defaults#otterdog-defaults.libsonnet")

    with pytest.raises(ValueError):
        parse_template_url("https://github.com/EclipseFdn/otterdog-defaults")


def test_is_ghsa_repo():
    assert is_ghsa_repo("name") is False
    assert is_ghsa_repo("name-wqjm-x66q-r2c6") is False
    assert is_ghsa_repo("jiro-ghsa-wqjm-x66q-r2c6") is True


def test_deep_merge_dict():
    src = {
        "first": {"Matt": 10, "Arnie": 2},
        "second": {"Peter": 2},
    }
    dst = {
        "first": {"Matt": 1},
        "third": {"Maria": 3},
    }

    assert deep_merge_dict(src, dst) == {
        "first": {"Matt": 10, "Arnie": 2},
        "second": {"Peter": 2},
        "third": {"Maria": 3},
    }
